# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Dict, Any

from mautrix.bridge import Bridge
from mautrix.types import RoomID, UserID

from .version import version, linkified_version
from .config import Config
from .db import upgrade_table, init as init_db
from .matrix import MatrixHandler
from .user import User
from .portal import Portal
from .puppet import Puppet
from .web import ProvisioningAPI
from . import commands


class TwitterBridge(Bridge):
    module = "mautrix_twitter"
    name = "mautrix-twitter"
    command = "python -m mautrix-twitter"
    description = "A Matrix-Twitter DM puppeting bridge."
    repo_url = "https://github.com/mautrix/twitter"
    version = version
    markdown_version = linkified_version
    config_class = Config
    matrix_class = MatrixHandler
    upgrade_table = upgrade_table

    matrix: MatrixHandler
    config: Config
    provisioning_api: ProvisioningAPI

    def prepare_db(self) -> None:
        super().prepare_db()
        init_db(self.db)

    def prepare_bridge(self) -> None:
        super().prepare_bridge()
        cfg = self.config["appservice.provisioning"]
        self.provisioning_api = ProvisioningAPI(cfg["shared_secret"])
        self.az.app.add_subapp(cfg["prefix"], self.provisioning_api.app)

    async def start(self) -> None:
        self.add_startup_actions(User.init_cls(self))
        self.add_startup_actions(Puppet.init_cls(self))
        Portal.init_cls(self)
        if self.config["bridge.resend_bridge_info"]:
            self.add_startup_actions(self.resend_bridge_info())
        await super().start()

    def prepare_stop(self) -> None:
        self.add_shutdown_actions(user.stop() for user in User.by_twid.values())
        for puppet in Puppet.by_custom_mxid.values():
            puppet.stop()

    async def resend_bridge_info(self) -> None:
        self.config["bridge.resend_bridge_info"] = False
        self.config.save()
        self.log.info("Re-sending bridge info state event to all portals")
        async for portal in Portal.all_with_room():
            await portal.update_bridge_info()
        self.log.info("Finished re-sending bridge info state events")

    async def get_user(self, user_id: UserID, create: bool = True) -> User:
        return await User.get_by_mxid(user_id, create=create)

    async def get_portal(self, room_id: RoomID) -> Portal:
        return await Portal.get_by_mxid(room_id)

    async def get_puppet(self, user_id: UserID, create: bool = False) -> Puppet:
        return await Puppet.get_by_mxid(user_id, create=create)

    async def get_double_puppet(self, user_id: UserID) -> Puppet:
        return await Puppet.get_by_custom_mxid(user_id)

    def is_bridge_ghost(self, user_id: UserID) -> bool:
        return bool(Puppet.get_id_from_mxid(user_id))

    async def count_logged_in_users(self) -> int:
        return len([user for user in User.by_twid.values() if user.twid])

    async def manhole_global_namespace(self, user_id: UserID) -> Dict[str, Any]:
        return {
            **await super().manhole_global_namespace(user_id),
            "User": User,
            "Portal": Portal,
            "Puppet": Puppet,
        }


TwitterBridge().run()



# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from asyncpg import Connection

from mautrix.util.async_db import UpgradeTable

upgrade_table = UpgradeTable()


@upgrade_table.register(description="Initial revision")
async def upgrade_v1(conn: Connection) -> None:
    await conn.execute("CREATE TYPE twitter_conv_type AS ENUM ('ONE_TO_ONE', 'GROUP_DM')")
    await conn.execute("""CREATE TABLE portal (
        twid        VARCHAR(255),
        receiver    BIGINT,
        conv_type   twitter_conv_type NOT NULL,
        other_user  BIGINT,
        mxid        VARCHAR(255),
        name        VARCHAR(255),
        encrypted   BOOLEAN NOT NULL DEFAULT false,

        PRIMARY KEY (twid, receiver)
    )""")
    await conn.execute("""CREATE TABLE "user" (
        mxid        VARCHAR(255) PRIMARY KEY,
        twid        BIGINT,
        auth_token  VARCHAR(255),
        csrf_token  VARCHAR(255),
        poll_cursor VARCHAR(255),
        notice_room VARCHAR(255)
    )""")
    await conn.execute("""CREATE TABLE puppet (
        twid      BIGINT PRIMARY KEY,
        name      VARCHAR(255),
        photo_url VARCHAR(255),
        photo_mxc VARCHAR(255),

        is_registered BOOLEAN NOT NULL DEFAULT false,

        custom_mxid  VARCHAR(255),
        access_token TEXT,
        next_batch   VARCHAR(255)
    )""")
    await conn.execute("""CREATE TABLE user_portal (
        "user"          BIGINT,
        portal          VARCHAR(255),
        portal_receiver BIGINT,
        in_community    BOOLEAN NOT NULL DEFAULT false,

        FOREIGN KEY (portal, portal_receiver) REFERENCES portal(twid, receiver)
            ON UPDATE CASCADE ON DELETE CASCADE
    )""")
    await conn.execute("""CREATE TABLE message (
        mxid     VARCHAR(255) NOT NULL,
        mx_room  VARCHAR(255) NOT NULL,
        twid     BIGINT,
        receiver BIGINT,

        PRIMARY KEY (twid, receiver),
        UNIQUE (mxid, mx_room)
    )""")
    await conn.execute("CREATE TYPE twitter_reaction_key AS ENUM ('funny', 'surprised', 'sad',"
                       "                                          'like', 'excited', 'agree',"
                       "                                          'disagree')")
    await conn.execute("""CREATE TABLE reaction (
        mxid        VARCHAR(255) NOT NULL,
        mx_room     VARCHAR(255) NOT NULL,
        tw_msgid    BIGINT,
        tw_receiver BIGINT,
        tw_sender   BIGINT,
        reaction    twitter_reaction_key NOT NULL,

        PRIMARY KEY (tw_msgid, tw_receiver, tw_sender),
        FOREIGN KEY (tw_msgid, tw_receiver) REFERENCES message(twid, receiver)
            ON DELETE CASCADE ON UPDATE CASCADE,
        UNIQUE (mxid, mx_room)
    )""")


@upgrade_table.register(description="Add double-puppeting base_url to puppet table")
async def upgrade_v2(conn: Connection) -> None:
    await conn.execute("ALTER TABLE puppet ADD COLUMN base_url TEXT")



# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Any, List, NamedTuple
import os

from mautrix.types import UserID
from mautrix.client import Client
from mautrix.util.config import ConfigUpdateHelper, ForbiddenKey, ForbiddenDefault
from mautrix.bridge.config import BaseBridgeConfig

Permissions = NamedTuple("Permissions", user=bool, admin=bool, level=str)


class Config(BaseBridgeConfig):
    def __getitem__(self, key: str) -> Any:
        try:
            return os.environ[f"MAUTRIX_TWITTER_{key.replace('.', '_').upper()}"]
        except KeyError:
            return super().__getitem__(key)

    @property
    def forbidden_defaults(self) -> List[ForbiddenDefault]:
        return [
            *super().forbidden_defaults,
            ForbiddenDefault("appservice.database", "postgres://username:password@hostname/db"),
            ForbiddenDefault("bridge.permissions", ForbiddenKey("example.com")),
        ]

    def do_update(self, helper: ConfigUpdateHelper) -> None:
        super().do_update(helper)
        copy, copy_dict, base = helper

        copy("homeserver.asmux")

        copy("appservice.provisioning.enabled")
        copy("appservice.provisioning.prefix")
        copy("appservice.provisioning.shared_secret")
        if base["appservice.provisioning.shared_secret"] == "generate":
            base["appservice.provisioning.shared_secret"] = self._new_token()

        copy("appservice.community_id")

        copy("metrics.enabled")
        copy("metrics.listen_port")

        copy("bridge.username_template")
        copy("bridge.displayname_template")

        copy("bridge.displayname_max_length")

        copy("bridge.initial_conversation_sync")
        copy("bridge.sync_with_custom_puppets")
        copy("bridge.sync_direct_chat_list")
        copy("bridge.double_puppet_server_map")
        copy("bridge.double_puppet_allow_discovery")
        if self["bridge.login_shared_secret"]:
            base["bridge.login_shared_secret_map"] = {
                base["homeserver.domain"]: self["bridge.login_shared_secret"]
            }
        else:
            copy("bridge.login_shared_secret_map")
        copy("bridge.federate_rooms")
        copy("bridge.backfill.invite_own_puppet")
        copy("bridge.backfill.initial_limit")
        copy("bridge.backfill.disable_notifications")
        copy("bridge.encryption.allow")
        copy("bridge.encryption.default")
        copy("bridge.encryption.key_sharing.allow")
        copy("bridge.encryption.key_sharing.require_cross_signing")
        copy("bridge.encryption.key_sharing.require_verification")
        copy("bridge.private_chat_portal_meta")
        copy("bridge.delivery_receipts")
        copy("bridge.delivery_error_reports")
        copy("bridge.temporary_disconnect_notices")
        copy("bridge.disable_bridge_notices")
        copy("bridge.error_sleep")
        copy("bridge.max_poll_errors")
        copy("bridge.resend_bridge_info")

        copy("bridge.command_prefix")

        copy_dict("bridge.permissions")

    def _get_permissions(self, key: str) -> Permissions:
        level = self["bridge.permissions"].get(key, "")
        admin = level == "admin"
        user = level == "user" or admin
        return Permissions(user, admin, level)

    def get_permissions(self, mxid: UserID) -> Permissions:
        permissions = self["bridge.permissions"]
        if mxid in permissions:
            return self._get_permissions(mxid)

        _, homeserver = Client.parse_user_id(mxid)
        if homeserver in permissions:
            return self._get_permissions(homeserver)

        return self._get_permissions("*")



# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Union
import html

from mautwitdm.types import MessageData, MessageEntityURL, MessageEntityUserMention
from mautrix.types import TextMessageEventContent, MessageType, Format

from . import puppet as pu

MessageEntity = Union[MessageEntityURL, MessageEntityUserMention]


async def twitter_to_matrix(message: MessageData) -> TextMessageEventContent:
    content = TextMessageEventContent(msgtype=MessageType.TEXT, body=message.text,
                                      format=Format.HTML, formatted_body=message.text)
    for entity in reversed(message.entities.all) if message.entities else []:
        start, end = entity.indices
        if isinstance(entity, MessageEntityURL):
            content.body = content.body[:start] + entity.expanded_url + content.body[end:]
            content.formatted_body = (f'{content.formatted_body[:start]}'
                                      f'<a href="{entity.expanded_url}">{entity.display_url}</a>'
                                      f'{content.formatted_body[end:]}')
        elif isinstance(entity, MessageEntityUserMention):
            puppet = await pu.Puppet.get_by_twid(entity.id, create=False)
            if puppet:
                user_url = f"https://matrix.to/#/{puppet.mxid}"
                content.formatted_body = (f'{content.formatted_body[:start]}'
                                          f'<a href="{user_url}">{puppet.name or entity.name}</a>'
                                          f'{content.formatted_body[end:]}')
        else:
            # Get the sigil (# or $) from the body
            text = content.formatted_body[start:end][0] + entity.text
            content.formatted_body = (f'{content.formatted_body[:start]}'
                                      f'<font color="#0000ff">{text}</font>'
                                      f'{content.formatted_body[end:]}')
    if content.formatted_body == content.body:
        content.formatted_body = None
        content.format = None
    content.body = html.unescape(content.body)
    return content



# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, ClassVar, TYPE_CHECKING

from attr import dataclass

from mautwitdm.types import ReactionKey
from mautrix.types import RoomID, EventID
from mautrix.util.async_db import Database

fake_db = Database("") if TYPE_CHECKING else None


@dataclass
class Reaction:
    db: ClassVar[Database] = fake_db

    mxid: EventID
    mx_room: RoomID
    tw_msgid: int
    tw_receiver: int
    tw_sender: int
    reaction: ReactionKey

    async def insert(self) -> None:
        q = ("INSERT INTO reaction (mxid, mx_room, tw_msgid, tw_receiver, tw_sender, reaction) "
             "VALUES ($1, $2, $3, $4, $5, $6)")
        await self.db.execute(q, self.mxid, self.mx_room, self.tw_msgid, self.tw_receiver,
                              self.tw_sender, self.reaction.value)

    async def edit(self, mx_room: RoomID, mxid: EventID, reaction: ReactionKey) -> None:
        await self.db.execute("UPDATE reaction SET mxid=$1, mx_room=$2, reaction=$3 "
                              "WHERE tw_msgid=$4 AND tw_receiver=$5 AND tw_sender=$6",
                              mxid, mx_room, reaction.value, self.tw_msgid, self.tw_receiver,
                              self.tw_sender)

    async def delete(self) -> None:
        q = "DELETE FROM reaction WHERE tw_msgid=$1 AND tw_receiver=$2 AND tw_sender=$3"
        await self.db.execute(q, self.tw_msgid, self.tw_receiver, self.tw_sender)

    @classmethod
    async def get_by_mxid(cls, mxid: EventID, mx_room: RoomID) -> Optional['Reaction']:
        q = ("SELECT mxid, mx_room, tw_msgid, tw_receiver, tw_sender, reaction "
             "FROM reaction WHERE mxid=$1 AND mx_room=$2")
        row = await cls.db.fetchrow(q, mxid, mx_room)
        if not row:
            return None
        data = {**row}
        return cls(reaction=ReactionKey(data.pop("reaction")), **data)

    @classmethod
    async def get_by_twid(cls, tw_msgid: int, tw_receiver: int, tw_sender: int,
                          ) -> Optional['Reaction']:
        q = ("SELECT mxid, mx_room, tw_msgid, tw_receiver, tw_sender, reaction "
             "FROM reaction WHERE tw_msgid=$1 AND tw_sender=$2 AND tw_receiver=$3")
        row = await cls.db.fetchrow(q, tw_msgid, tw_sender, tw_receiver)
        if not row:
            return None
        data = {**row}
        return cls(reaction=ReactionKey(data.pop("reaction")), **data)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, Union, List

from attr import dataclass

from mautrix.types import SerializableAttrs


@dataclass
class DMTypingEvent(SerializableAttrs):
    user_id: str
    conversation_id: str


@dataclass
class DMUpdateEvent(SerializableAttrs):
    user_id: str
    conversation_id: str


@dataclass
class SubscriptionError(SerializableAttrs):
    topic: str
    code: int
    message: str


@dataclass
class SubscriptionsEvent(SerializableAttrs):
    errors: List[SubscriptionError]


StreamEventType = Union[SubscriptionsEvent, DMTypingEvent, DMUpdateEvent]


@dataclass
class StreamEvent(SerializableAttrs):
    dm_typing: Optional[DMTypingEvent] = None
    dm_update: Optional[DMUpdateEvent] = None
    subscriptions: Optional[SubscriptionsEvent] = None

    @property
    def all_types(self) -> List[StreamEventType]:
        items = (self.dm_typing, self.dm_update)
        return [item for item in items
                if item is not None]



# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from mautrix.bridge.commands import HelpSection, command_handler
from .typehint import CommandEvent

SECTION_AUTH = HelpSection("Authentication", 10, "")


@command_handler(needs_auth=False, management_only=True,
                 help_section=SECTION_AUTH, help_text="Log in to Twitter manually")
async def login_cookie(evt: CommandEvent) -> None:
    if evt.sender.client:
        await evt.reply("You're already logged in")
        return
    evt.sender.command_status = {
        "action": "Login",
        "room_id": evt.room_id,
        "next": enter_login_cookies,
        "auth_token": None,
    }
    await evt.reply(
        "1. Log in to [Twitter](https://www.twitter.com/) in a private/incognito window.\n"
        "2. Press `F12` to open developer tools.\n"
        "3. Select the \"Application\" (Chrome) or \"Storage\" (Firefox) tab.\n"
        "4. In the sidebar, expand \"Cookies\" and select `https://twitter.com`.\n"
        "5. In the cookie list, find the `auth_token` row and double click on the value"
        r", then copy the value and send it here.")


async def enter_login_cookies(evt: CommandEvent) -> None:
    if not evt.sender.command_status["auth_token"]:
        if len(evt.args) == 0:
            await evt.reply("Please enter the value of the `auth_token` cookie, or use "
                            "the `cancel` command to cancel.")
            return
        evt.sender.command_status["auth_token"] = evt.args[0]
        await evt.reply("Now do the last step again, but find the value of the `ct0` row instead. "
                        "Before you send the value, close the private window.")
        return
    if len(evt.args) == 0:
        await evt.reply("Please enter the value of the `ct0` cookie, or use "
                        "the `cancel` command to cancel.")
        return

    try:
        await evt.sender.connect(auth_token=evt.sender.command_status["auth_token"],
                                 csrf_token=evt.args[0])
    except Exception as e:
        evt.sender.command_status = None
        await evt.reply(f"Failed to log in: {e}")
        evt.log.exception("Failed to log in")
        return

    await evt.reply(f"Successfully logged in as @{evt.sender.username}")
    evt.sender.command_status = None


@command_handler(needs_auth=True, help_section=SECTION_AUTH, help_text="Disconnect the bridge from"
                                                                       "your Twitter account")
async def logout(evt: CommandEvent) -> None:
    await evt.sender.logout()
    await evt.reply("Successfully logged out")



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, Union, TYPE_CHECKING
from uuid import UUID

from yarl import URL

from .types import ReactionKey, SendResponse, FetchConversationResponse
from .errors import check_error

if TYPE_CHECKING:
    from .twitter import TwitterAPI


class Conversation:
    api: 'TwitterAPI'
    id: str

    def __init__(self, api: 'TwitterAPI', id: str) -> None:
        self.api = api
        self.id = id

    @property
    def api_url(self) -> URL:
        """The base URL for API requests related to this conversation."""
        return self.api.dm_url / "conversation" / self.id

    async def mark_read(self, last_read_event_id: Union[int, str]) -> None:
        """Mark the conversation as read, up to the given event ID."""
        req = {"conversationId": self.id,
               "last_read_event_id": str(last_read_event_id)}
        url = self.api_url / "mark_read.json"
        async with self.api.http.post(url, headers=self.api.headers, data=req) as resp:
            await check_error(resp)

    async def mark_typing(self) -> None:
        """Send a typing notification. This request should be repeated every 2-3 seconds."""
        async with self.api.http.post(self.api_url / "typing.json", headers=self.api.headers) as r:
            await check_error(r)

    async def accept(self) -> None:
        """Accept the conversation (when DMing users who you don't follow)."""
        async with self.api.http.post(self.api_url / "accept.json", headers=self.api.headers) as r:
            await check_error(r)

    async def send(self, text: str, media_id: Optional[Union[str, int]] = None,
                   request_id: Optional[Union[UUID, str]] = None) -> SendResponse:
        """
        Send a message to this conversation.

        Args:
            text: The text to send. May be an empty string if only sending media.
            media_id: The media ID to send. Use :meth:`TwitterUploader.upload` to upload media and
                get a media ID.
            request_id: The transaction ID for this request. It will be included in the message
                when polling and can be used for deduplication. If not provided, one will be
                automatically created using :meth:`TwitterAPI.new_request_id`

        Returns:
            The send response from the server.
        """
        data = {
            **self.api.poll_params,
            "text": text,
            "conversation_id": self.id,
            "recipient_ids": "false",
            "request_id": str(request_id or self.api.new_request_id()),
        }
        url = self.api.dm_url / "new.json"
        if media_id:
            data["media_id"] = str(media_id)
        async with self.api.http.post(url, data=data, headers=self.api.headers) as resp:
            resp_data = await check_error(resp)
            return SendResponse.deserialize(resp_data)

    async def react(self, message_id: Union[str, int], key: ReactionKey) -> None:
        """
        React to a message. Reacting to the same message multiple times will override earlier
        reactions.

        Args:
            message_id: The message ID to react to.
            key: The reaction itself.
        """
        url = (self.api.dm_url / "reaction" / "new.json").with_query({
            "reaction_key": str(key),
            "conversation_id": self.id,
            "dm_id": str(message_id),
        })
        async with self.api.http.post(url, headers=self.api.headers) as resp:
            await check_error(resp)

    async def delete_reaction(self, message_id: Union[str, int], key: ReactionKey) -> None:
        """
        Delete an earlier reaction.

        Args:
            message_id: The message ID to react to.
            key: The reaction itself.
        """
        url = (self.api.dm_url / "reaction" / "delete.json").with_query({
            "reaction_key": str(key),
            "conversation_id": self.id,
            "dm_id": str(message_id),
        })
        async with self.api.http.post(url, headers=self.api.headers) as resp:
            await check_error(resp)

    async def fetch(self, max_id: Optional[str] = None, include_info: bool = True
                    ) -> FetchConversationResponse:
        """
        Fetch the conversation metadata and message history.

        Args:
            max_id: The maximum message ID to fetch.
            include_info: Whether or not to include conversation metadata in the response.

        Returns:
            The requested metadata and message history.
        """
        query = {
            **self.api.full_state_params,
            "context": "FETCH_DM_CONVERSATION",
        }
        if include_info:
            query["include_conversation_info"] = "true"
        req = (self.api.dm_url / "conversation" / f"{self.id}.json").with_query(query)
        if max_id:
            req = req.update_query({"max_id": max_id})
        async with self.api.http.get(req, headers=self.api.headers) as resp:
            resp_data = await check_error(resp)
        return FetchConversationResponse.deserialize(resp_data["conversation_timeline"])



# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from mautrix.bridge.commands import HelpSection, command_handler
from .typehint import CommandEvent

SECTION_CONNECTION = HelpSection("Connection management", 15, "")


@command_handler(needs_auth=False, management_only=True, help_section=SECTION_CONNECTION,
                 help_text="Mark this room as your bridge notice room")
async def set_notice_room(evt: CommandEvent) -> None:
    evt.sender.notice_room = evt.room_id
    await evt.sender.update()
    await evt.reply("This room has been marked as your bridge notice room")


@command_handler(needs_auth=False, management_only=True, help_section=SECTION_CONNECTION,
                 help_text="Check if you're logged into Twitter")
async def ping(evt: CommandEvent) -> None:
    if not await evt.sender.is_logged_in():
        await evt.reply("You're not logged into Twitter")
        return
    user_info = await evt.sender.get_info()
    await evt.reply(f"You're logged in as {user_info.name} "
                    f"([@{evt.sender.username}](https://twitter.com/{evt.sender.username}), "
                    f"user ID: {evt.sender.twid})")


@command_handler(needs_auth=True, management_only=False, help_section=SECTION_CONNECTION,
                 help_text="Synchronize portals")
async def sync(evt: CommandEvent) -> None:
    await evt.sender.sync()
    await evt.reply("Synchronization complete")


@command_handler(needs_auth=True, management_only=False, help_section=SECTION_CONNECTION,
                 help_text="Start the Twitter poll task")
async def start_polling(evt: CommandEvent) -> None:
    if evt.sender.client.is_polling():
        await evt.reply("You already have a poll loop. If you want to restart it, "
                        "use `$cmdprefix+sp stop-polling` first.")
    else:
        evt.sender.client.start_polling()
        await evt.reply("Started poll task")


@command_handler(needs_auth=True, management_only=False, help_section=SECTION_CONNECTION,
                 help_text="Stop the Twitter poll task")
async def stop_polling(evt: CommandEvent) -> None:
    if evt.sender.client.is_polling():
        evt.sender.client.stop_polling()
        await evt.reply("Stopped poll task")
    else:
        await evt.reply("You don't have a running poll task. "
                        "You can start one with `$cmdprefix+sp start-polling`.")



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import AsyncGenerator, Set, Dict
import logging
import asyncio
import json
import io

from aiohttp import ClientSession
from yarl import URL

from .types import StreamEvent
from .dispatcher import TwitterDispatcher
from .errors import check_error


class TwitterStreamer(TwitterDispatcher):
    """
    This class handles receiving live events like typing notifications
    via ``/live_pipeline/events``.
    """
    pipeline_url = URL("https://api.twitter.com/live_pipeline/events")
    pipeline_update_url = URL("https://api.twitter.com/1.1/live_pipeline/update_subscriptions")

    log: logging.Logger
    loop: asyncio.AbstractEventLoop
    http: ClientSession
    headers: Dict[str, str]
    user_agent: str

    topics: Set[str]
    _stream_task: asyncio.Task

    async def _stream(self) -> AsyncGenerator[StreamEvent, None]:
        url = self.pipeline_url.with_query(topics=",".join(self.topics))
        headers = {
            "User-Agent": self.user_agent,
            "Accept": "text/event-stream",
            "Accept-Language": "en-US,en;q=0.5",
            "DNT": "1",
            "Origin": "https://twitter.com",
            "Referer": "https://twitter.com/messages",
            "Pragma": "no-cache",
        }
        empty_bytes = b""
        data_prefix = b"data: "
        chunk_separator = b"\n\n"
        empty_chunk = b":\n\n"
        async with self.http.get(url, headers=headers) as resp:
            buffer = io.BytesIO()
            partial = False
            while True:
                chunk, end = await resp.content.readchunk()
                if not end and chunk == empty_bytes:
                    break
                if not chunk.endswith(chunk_separator):
                    buffer.write(chunk)
                    partial = True
                    continue
                if partial:
                    chunk = buffer.getvalue() + chunk
                    buffer = io.BytesIO()
                    partial = False
                if chunk == empty_chunk:
                    continue
                elif chunk.startswith(data_prefix):
                    data = json.loads(chunk[len(data_prefix):-len(chunk_separator)])
                    yield StreamEvent.deserialize(data["payload"])

    async def update_topics(self, subscribe: Set[str], unsubscribe: Set[str]) -> None:
        """
        Update the topics the client is subscribed to.

        Args:
            subscribe: The list of topics to subscribe to.
            unsubscribe: The list of topics to unsubscribe from.
        """
        subscribe = subscribe - self.topics
        unsubscribe = unsubscribe & self.topics
        self.topics = self.topics - unsubscribe | subscribe
        if self._stream_task is None or self._stream_task.done():
            self.log.debug("Not sending update_subscriptions request: no ongoing stream task")
            return
        url = self.pipeline_update_url
        req = {"sub_topics": ",".join(subscribe), "unsub_topics": ",".join(unsubscribe)}
        async with self.http.post(url, data=req, headers=self.headers) as resp:
            resp_data = await check_error(resp)
        event = StreamEvent.deserialize(resp_data)
        for event_type in event.all_types:
            await self.dispatch(event_type)

    async def stream_forever(self, raise_exceptions: bool = True) -> None:
        """
        Stream for events from the server forever. Events are dispatched using :meth:`dispatch`,
        which means handlers can be added with :meth:`add_handler`.

        You can set which updates to listen to by setting :attribute:`topics` before calling this,
        or by calling :meth:`update_topics` while streaming.

        Args:
            raise_exceptions: Whether or not errors should be raised after logging. If set to
                ``False``, errors will be logged, and the stream will retry in 10 seconds.
                :class:`asyncio.CancelledError` will not be raised and will simply return.
        """
        while True:
            try:
                async for event in self._stream():
                    for event_type in event.all_types:
                        await self.dispatch(event_type)
            except asyncio.CancelledError:
                self.log.debug("Streaming stopped")
                break
            except Exception:
                self.log.exception("Error while streaming events")
                if raise_exceptions:
                    raise
                await asyncio.sleep(10)

    def start_streaming(self) -> asyncio.Task:
        """
        Start polling forever in the background. This calls :meth:`stream_forever` and puts it in
        an asyncio Task. The task is stored so it can be cancelled with :meth:`stop_streaming`.

        Returns:
            The created asyncio task.
        """
        self.log.debug("Starting streaming task")
        self._stream_task = asyncio.create_task(self.stream_forever())
        return self._stream_task

    def stop_streaming(self) -> None:
        """Stop the ongoing stream task. Any ongoing handlers will also be cancelled."""
        if self._stream_task:
            self.log.debug("Cancelling ongoing stream task")
            self._stream_task.cancel()



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import NewType
from datetime import datetime, timezone

from mautrix.types import deserializer, serializer

StringTimestamp = NewType('StringTimestamp', datetime)


@deserializer(StringTimestamp)
def deserialize_str_timestamp(val: str) -> StringTimestamp:
    return StringTimestamp(datetime.utcfromtimestamp(int(val) / 1000))


@serializer(StringTimestamp)
def serialize_str_timestamp(val: StringTimestamp) -> str:
    return str(val.timestamp() * 1000)


StringDateTime = NewType('StringDateTime', datetime)


@deserializer(StringDateTime)
def deserialize_str_datetime(val: str) -> StringDateTime:
    return StringDateTime(datetime.strptime(val, "%a %b %d %H:%M:%S +0000 %Y")
                          .replace(tzinfo=timezone.utc))


@serializer(StringDateTime)
def serialize_str_datetime(val: StringDateTime) -> str:
    return str(val.strftime("%a %b %d %H:%M:%S +0000 %Y"))



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict
import logging
import asyncio
import math

from aiohttp import ClientSession, MultipartWriter
from yarl import URL

from .types import MediaUploadResponse
from .errors import check_error


class TwitterUploader:
    upload_url: URL = URL("https://upload.twitter.com/i/media/upload.json")

    http: ClientSession
    log: logging.Logger
    headers: Dict[str, str]

    async def _wait_processing(self, media_id: str, wait_requests: int = 1) -> MediaUploadResponse:
        query_req = self.upload_url.with_query({
            "command": "STATUS",
            "media_id": media_id
        })
        while True:
            await asyncio.sleep(wait_requests)
            async with self.http.get(query_req, headers=self.headers) as resp:
                data = await check_error(resp)
            state = data["processing_info"]["state"]
            if state == "succeeded":
                self.log.debug(f"Server completed processing of {media_id}")
                return MediaUploadResponse.deserialize(data)
            elif state == "in_progress":
                progress = data["processing_info"]["progress_percent"]
                wait_requests = data["processing_info"].get("check_after_secs", wait_requests)
                self.log.debug(f"Upload of {media_id} at {progress} %, "
                               f"re-checking after {wait_requests} seconds")
            else:
                raise RuntimeError(f"Unknown state {state}")

    async def _upload_data(self, media_id: str, data: bytes) -> MediaUploadResponse:
        max_size = 2 ** 17
        base_upload_req = self.upload_url.with_query({
            "command": "APPEND",
            "media_id": media_id,
        })
        for i in range(math.ceil(len(data) / max_size)):
            multipart_data = MultipartWriter("form-data")
            part = multipart_data.append(data[i * max_size: (i + 1) * max_size])
            part.set_content_disposition("form-data", name="media", filename="blob")
            req = base_upload_req.update_query({"segment_index": i})
            async with self.http.post(req, data=multipart_data, headers=self.headers) as resp:
                await check_error(resp)
                self.log.debug(f"Uploaded segment {i} of {media_id}")
        finalize_req = self.upload_url.with_query({
            "command": "FINALIZE",
            "media_id": media_id,
        })
        async with self.http.post(finalize_req, headers=self.headers) as resp:
            resp_data = await check_error(resp)
        processing_info = resp_data.get("processing_info", {})
        if processing_info.get("state", None) == "pending":
            self.log.debug(f"Finished uploading {media_id}, but server is still processing it")
            check_after = processing_info.get("check_after_secs", 1)
            return await self._wait_processing(media_id, check_after)
        self.log.debug(f"Finished uploading {media_id}")
        return MediaUploadResponse.deserialize(resp_data)

    async def upload(self, data: bytes = None, url: str = None, mime_type: str = None
                     ) -> MediaUploadResponse:
        if mime_type == "image/gif":
            category = "dm_gif"
            size_limit = 15 * 1024 * 1024
        elif mime_type.startswith("image/"):
            category = "dm_image"
            size_limit = 5 * 1024 * 1024
        elif mime_type.startswith("video/"):
            category = "dm_video"
            size_limit = 15 * 1024 * 1024
        else:
            raise NotImplementedError(f"Unsupported mime type {mime_type}")
        if len(data) > size_limit:
            raise NotImplementedError("File too big")
        init_req = {
            "command": "INIT",
            "media_type": mime_type,
            "media_category": category,
        }
        if data is not None:
            init_req["total_bytes"] = len(data)
        elif url is not None:
            init_req["source_url"] = url
        else:
            raise ValueError("Either data bytes or url must be provided")
        init_url = self.upload_url.with_query(init_req)
        async with self.http.post(init_url, headers=self.headers) as resp:
            resp_data = await check_error(resp)
        media_id = resp_data["media_id_string"]
        self.log.debug(f"Started upload, got media ID {media_id}")

        if url is not None:
            processing_info = resp_data.get("processing_info", {})
            if processing_info.get("state") == "succeeded":
                return MediaUploadResponse.deserialize(data)
            else:
                check_after = processing_info.get("check_after_secs", 1)
                return await self._wait_processing(media_id, check_after)
        else:
            return await self._upload_data(media_id, data)



# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, ClassVar, List, TYPE_CHECKING

from attr import dataclass

from mautrix.types import UserID, RoomID
from mautrix.util.async_db import Database

fake_db = Database("") if TYPE_CHECKING else None


@dataclass
class User:
    db: ClassVar[Database] = fake_db

    mxid: UserID
    twid: Optional[int]
    auth_token: Optional[str]
    csrf_token: Optional[str]
    poll_cursor: Optional[str]
    notice_room: Optional[RoomID]

    async def insert(self) -> None:
        q = ('INSERT INTO "user" (mxid, twid, auth_token, csrf_token, poll_cursor, notice_room) '
             'VALUES ($1, $2, $3, $4, $5, $6)')
        await self.db.execute(q, self.mxid, self.twid, self.auth_token, self.csrf_token,
                              self.poll_cursor, self.notice_room)

    async def update(self) -> None:
        await self.db.execute('UPDATE "user" SET twid=$2, auth_token=$3, csrf_token=$4,'
                              '                  poll_cursor=$5, notice_room=$6 '
                              'WHERE mxid=$1', self.mxid, self.twid, self.auth_token,
                              self.csrf_token, self.poll_cursor, self.notice_room)

    @classmethod
    async def get_by_mxid(cls, mxid: UserID) -> Optional['User']:
        q = ("SELECT mxid, twid, auth_token, csrf_token, poll_cursor, notice_room "
             'FROM "user" WHERE mxid=$1')
        row = await cls.db.fetchrow(q, mxid)
        if not row:
            return None
        return cls(**row)

    @classmethod
    async def get_by_twid(cls, twid: int) -> Optional['User']:
        q = ("SELECT mxid, twid, auth_token, csrf_token, poll_cursor, notice_room "
             'FROM "user" WHERE twid=$1')
        row = await cls.db.fetchrow(q, twid)
        if not row:
            return None
        return cls(**row)

    @classmethod
    async def all_logged_in(cls) -> List['User']:
        q = ("SELECT mxid, twid, auth_token, csrf_token, poll_cursor, notice_room "
             'FROM "user" WHERE twid IS NOT NULL AND auth_token IS NOT NULL')
        rows = await cls.db.fetch(q)
        return [cls(**row) for row in rows]



# Copyright (c) 2021 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import List, Tuple, Union

from attr import dataclass

from mautrix.types import SerializableAttrs


@dataclass
class MessageEntityURL(SerializableAttrs):
    url: str
    expanded_url: str
    display_url: str
    indices: Tuple[int, int]


@dataclass
class MessageEntityUserMention(SerializableAttrs):
    screen_name: str
    name: str
    id: int
    id_str: str
    indices: Tuple[int, int]


@dataclass
class MessageEntitySimple(SerializableAttrs):
    text: str
    indices: Tuple[int, int]


@dataclass
class MessageEntities(SerializableAttrs):
    hashtags: List[MessageEntitySimple]
    symbols: List[MessageEntitySimple]
    user_mentions: List[MessageEntityUserMention]
    urls: List[MessageEntityURL]

    @property
    def all(self) -> List[Union[MessageEntitySimple, MessageEntityUserMention, MessageEntityURL]]:
        entities = self.hashtags + self.symbols + self.user_mentions + self.urls
        return sorted(entities, key=lambda entity: entity.indices[0])



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, Dict, Any, NamedTuple, List
from uuid import UUID, uuid1, getnode
from collections import defaultdict
from http.cookies import SimpleCookie
import logging
import asyncio

from aiohttp import ClientSession
from yarl import URL

from .types import User
from .conversation import Conversation
from .uploader import TwitterUploader
from .streamer import TwitterStreamer
from .poller import TwitterPoller
from .errors import TwitterError, check_error

Tokens = NamedTuple('Tokens', auth_token=str, csrf_token=str)
DownloadResp = NamedTuple('DownloadResp', data=bytes, mime_type=str)

twitter_com = URL("https://twitter.com/")


class TwitterAPI(TwitterUploader, TwitterStreamer, TwitterPoller):
    """The main entrypoint for using the internal Twitter DM API."""
    base_url: URL = URL("https://api.twitter.com/1.1")
    dm_url: URL = base_url / "dm"

    loop: asyncio.AbstractEventLoop
    http: ClientSession
    log: logging.Logger

    node_id: int
    active: bool
    user_agent: str

    def __init__(self, http: Optional[ClientSession] = None, log: Optional[logging.Logger] = None,
                 loop: Optional[asyncio.AbstractEventLoop] = None, node_id: Optional[int] = None
                 ) -> None:
        self.loop = loop or asyncio.get_event_loop()
        self.http = http or ClientSession(loop=self.loop)
        self.log = log or logging.getLogger("mautwitdm")
        self.node_id = node_id or getnode()
        self.poll_cursor = None
        self._poll_task = None
        self.dispatch_initial_resp = False
        self._handlers = defaultdict(lambda: [])
        self.active = True
        self._typing_in = None
        self.user_agent = ("Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:89.0) "
                           "Gecko/20100101 Firefox/89.0")
        self.skip_poll_wait = asyncio.Event()
        self.topics = set()

    def set_tokens(self, auth_token: str, csrf_token: str) -> None:
        """
        Set the authentication tokens. After this, use :meth:`get_user_identifier` to check if the
        auth is working correctly.

        Args:
            auth_token: The auth_token cookie value.
            csrf_token: The ct0 cookie/x-csrf-token header value.
        """
        cookie = SimpleCookie()
        cookie["auth_token"] = auth_token
        cookie["auth_token"].update({"domain": "twitter.com", "path": "/"})
        cookie["ct0"] = csrf_token
        cookie["ct0"].update({"domain": "twitter.com", "path": "/"})
        self.http.cookie_jar.update_cookies(cookie, twitter_com)

    def mark_typing(self, conversation_id: Optional[str]) -> None:
        """
        Mark the user as typing in the specified conversation. This will make the polling task call
        :meth:`Conversation.mark_typing` of the specified conversation after each poll.

        Args:
            conversation_id: The conversation where the user is typing, or ``None`` to stop typing.
        """
        self._typing_in = self.conversation(conversation_id)

    @property
    def tokens(self) -> Optional[Tokens]:
        cookies = self.http.cookie_jar.filter_cookies(URL("https://twitter.com/"))
        try:
            return Tokens(auth_token=cookies["auth_token"].value, csrf_token=cookies["ct0"].value)
        except KeyError:
            return None

    @property
    def headers(self) -> Dict[str, str]:
        """
        Get the headers to use with every request to Twitter.

        Returns:
            A key-value HTTP header list.
        """
        csrf_token = self.http.cookie_jar.filter_cookies(twitter_com)["ct0"].value
        return {
            # Hardcoded authorization header from the web app
            "authorization": "Bearer AAAAAAAAAAAAAAAAAAAAANRILgAAAAAAnNwIzUejRCOuH5E6I8xnZz4puTs"
                             "%3D1Zv7ttfk8LF81IUq16cHjhLTvJu4FA33AGWWjCpTnA",
            "User-Agent": self.user_agent,
            "Accept": "*/*",
            "Accept-Language": "en-US,en;q=0.5",
            "DNT": "1",
            "Origin": "https://twitter.com",
            "Referer": "https://twitter.com/messages",
            "x-twitter-auth-type": "OAuth2Session",
            "x-twitter-client-language": "en",
            "x-twitter-active-user": "yes",
            "x-csrf-token": csrf_token,
        }

    async def download_media(self, url: str) -> DownloadResp:
        headers = {
            "Accept": "*/*",
            "DNT": "1",
            "Referer": "https://twitter.com/messages",
            "User-Agent": self.user_agent,
        }
        async with self.http.get(url, headers=headers) as resp:
            await check_error(resp)
            return DownloadResp(data=await resp.read(), mime_type=resp.headers["Content-Type"])

    def new_request_id(self) -> UUID:
        """
        Create a new request ID for DM send requests.

        Returns:
            A v1 UUID.
        """
        return uuid1(self.node_id)

    def conversation(self, id: str) -> Conversation:
        return Conversation(self, id)

    async def update_last_seen_event_id(self, last_seen_event_id: str) -> None:
        await self.http.post(self.dm_url / "update_last_seen_event_id.json",
                             data={"last_seen_event_id": last_seen_event_id,
                                   "trusted_last_seen_event_id": last_seen_event_id},
                             headers=self.headers)

    async def get_user_identifier(self) -> Optional[str]:
        async with self.http.post(self.base_url / "branch" / "init.json", json={},
                                  headers=self.headers) as resp:
            try:
                resp_data = await check_error(resp)
            except TwitterError as e:
                # Sometimes branch/init.json returns 38: countryCode parameter is missing.
                # It still checks auth and we don't actually need this user identifier,
                # so it might be safe to ignore
                if e.code == 38:
                    self.log.warning(f"Ignoring {e} in branch/init.json request")
                    return ""
                raise
            return resp_data.get("user_identifier", None)

    async def get_settings(self) -> Dict[str, Any]:
        """Get the account settings of the currently logged in account."""
        async with self.http.get(self.base_url / "account" / "settings.json",
                                 headers=self.headers) as resp:
            return await check_error(resp)

    async def lookup_users(self, user_ids: Optional[List[int]] = None,
                           usernames: Optional[List[str]] = None) -> List[User]:
        query = {"include_entities": "false", "tweet_mode": "extended"}
        if user_ids:
            query["user_id"] = ",".join(str(id) for id in user_ids)
        if usernames:
            query["screen_name"] = ",".join(usernames)
        req = (self.base_url / "users" / "lookup.json").with_query(query)
        async with self.http.get(req, headers=self.headers) as resp:
            resp_data = await check_error(resp)
            return [User.deserialize(user) for user in resp_data]



# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import List, Union, TYPE_CHECKING

from mautrix.bridge import BaseMatrixHandler
from mautrix.types import (Event, ReactionEvent, RoomID, EventID, UserID, ReactionEventContent,
                           RelationType, EventType, ReceiptEvent, TypingEvent, PresenceEvent,
                           RedactionEvent, SingleReceiptEventContent)

from .db import Message as DBMessage
from . import portal as po, user as u

if TYPE_CHECKING:
    from .__main__ import TwitterBridge


class MatrixHandler(BaseMatrixHandler):
    def __init__(self, bridge: 'TwitterBridge') -> None:
        prefix, suffix = bridge.config["bridge.username_template"].format(userid=":").split(":")
        homeserver = bridge.config["homeserver.domain"]
        self.user_id_prefix = f"@{prefix}"
        self.user_id_suffix = f"{suffix}:{homeserver}"

        super().__init__(bridge=bridge)

    async def send_welcome_message(self, room_id: RoomID, inviter: 'u.User') -> None:
        await super().send_welcome_message(room_id, inviter)
        if not inviter.notice_room:
            inviter.notice_room = room_id
            await inviter.update()
            await self.az.intent.send_notice(room_id, "This room has been marked as your "
                                                      "Twitter DM bridge notice room.")

    async def handle_leave(self, room_id: RoomID, user_id: UserID, event_id: EventID) -> None:
        portal = await po.Portal.get_by_mxid(room_id)
        if not portal:
            return

        user = await u.User.get_by_mxid(user_id, create=False)
        if not user:
            return

        await portal.handle_matrix_leave(user)

    @staticmethod
    async def handle_redaction(room_id: RoomID, user_id: UserID, event_id: EventID,
                               redaction_event_id: EventID) -> None:
        user = await u.User.get_by_mxid(user_id)
        if not user:
            return

        portal = await po.Portal.get_by_mxid(room_id)
        if not portal:
            return

        await portal.handle_matrix_redaction(user, event_id, redaction_event_id)

    @classmethod
    async def handle_reaction(cls, room_id: RoomID, user_id: UserID, event_id: EventID,
                              content: ReactionEventContent) -> None:
        if content.relates_to.rel_type != RelationType.ANNOTATION:
            cls.log.debug(f"Ignoring m.reaction event in {room_id} from {user_id} with unexpected "
                          f"relation type {content.relates_to.rel_type}")
            return
        user = await u.User.get_by_mxid(user_id)
        if not user:
            return

        portal = await po.Portal.get_by_mxid(room_id)
        if not portal:
            return

        await portal.handle_matrix_reaction(user, event_id, content.relates_to.event_id,
                                            content.relates_to.key)

    async def handle_read_receipt(self, user: 'u.User', portal: 'po.Portal', event_id: EventID,
                                  data: SingleReceiptEventContent) -> None:
        message = await DBMessage.get_by_mxid(event_id, portal.mxid)
        if not message:
            return
        user.log.debug(f"Marking messages in {portal.twid} read up to {message.twid}")
        await user.client.conversation(portal.twid).mark_read(message.twid)

    @staticmethod
    async def handle_typing(room_id: RoomID, typing: List[UserID]) -> None:
        # TODO implement
        pass

    async def handle_event(self, evt: Event) -> None:
        if evt.type == EventType.ROOM_REDACTION:
            evt: RedactionEvent
            await self.handle_redaction(evt.room_id, evt.sender, evt.redacts, evt.event_id)
        elif evt.type == EventType.REACTION:
            evt: ReactionEvent
            await self.handle_reaction(evt.room_id, evt.sender, evt.event_id, evt.content)

    async def handle_ephemeral_event(self, evt: Union[ReceiptEvent, PresenceEvent, TypingEvent]
                                     ) -> None:
        if evt.type == EventType.TYPING:
            await self.handle_typing(evt.room_id, evt.content.user_ids)
        else:
            await super().handle_ephemeral_event(evt)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, List, Dict, Tuple

from attr import dataclass

from mautrix.types import SerializableAttrs

from .user import User
from .message_entity import MessageEntities
from .util import StringDateTime


@dataclass
class ImageSizeInfo(SerializableAttrs):
    resize: str
    w: int
    h: int


@dataclass
class OriginalImageSizeInfo(SerializableAttrs):
    width: int
    height: int


@dataclass
class VideoVariant(SerializableAttrs):
    content_type: str
    url: str
    bitrate: Optional[int] = None


@dataclass
class VideoInfo(SerializableAttrs):
    aspect_ratio: Tuple[int, int]
    variants: List[VideoVariant]
    duration_millis: Optional[int] = None


@dataclass
class MessageAttachmentMedia(SerializableAttrs):
    id: int
    id_str: str
    indices: Tuple[int, int]
    media_url: str
    media_url_https: str
    url: str
    display_url: str
    expanded_url: str
    type: str
    sizes: Dict[str, ImageSizeInfo]
    original_info: OriginalImageSizeInfo
    video_info: Optional[VideoInfo] = None


@dataclass
class ImageBindingValue(SerializableAttrs):
    url: str
    width: int
    height: int


@dataclass
class CardBindingValue(SerializableAttrs):
    type: str
    string_value: Optional[str] = None
    image_value: Optional[ImageBindingValue] = None
    scribe_key: Optional[str] = None


@dataclass
class MessageAttachmentCard(SerializableAttrs):
    name: str
    url: str
    binding_values: Dict[str, CardBindingValue]
    users: Optional[Dict[str, User]] = None


@dataclass
class MessageEntityMedia(MessageAttachmentMedia, SerializableAttrs):
    indices: Optional[Tuple[int, int]] = None


@dataclass
class ExtendedMessageEntities(SerializableAttrs):
    media: List[MessageEntityMedia]


@dataclass
class TweetAttachmentStatus(SerializableAttrs):
    created_at: StringDateTime
    id: int
    id_str: str
    full_text: str
    truncated: bool
    display_text_range: Tuple[int, int]
    source: str
    user: User
    in_reply_to_status_id: Optional[int]
    in_reply_to_status_id_str: Optional[str]
    in_reply_to_user_id: Optional[int]
    in_reply_to_user_id_str: Optional[str]
    in_reply_to_screen_name: Optional[str]
    is_quote_status: bool
    retweet_count: int
    favorite_count: int
    reply_count: int
    quote_count: int
    favorited: bool
    retweeted: bool
    lang: Optional[str] = None
    supplemental_language: Optional[str] = None
    possibly_sensitive: Optional[bool] = None
    possibly_sensitive_editable: Optional[bool] = None
    entities: Optional[MessageEntities] = None
    extended_entities: Optional[ExtendedMessageEntities] = None


@dataclass
class MessageAttachmentTweet(SerializableAttrs):
    id: str
    url: str
    display_url: str
    expanded_url: str
    status: TweetAttachmentStatus
    indices: Optional[Tuple[int, int]] = None


@dataclass
class MessageAttachment(SerializableAttrs):
    photo: Optional[MessageAttachmentMedia] = None
    video: Optional[MessageAttachmentMedia] = None
    animated_gif: Optional[MessageAttachmentMedia] = None
    tweet: Optional[MessageAttachmentTweet] = None

    @property
    def media(self) -> Optional[MessageAttachmentMedia]:
        if self.video:
            return self.video
        if self.animated_gif:
            return self.animated_gif
        if self.photo:
            return self.photo



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Any
import json
import time

from aiohttp import ClientResponse, ContentTypeError
from multidict import CIMultiDictProxy


class TwitterError(Exception):
    code: int
    message: str

    def __init__(self, code: int, message: str) -> None:
        super().__init__(f"{code}: {message}")
        self.code = code
        self.message = message


class TwitterAuthError(TwitterError):
    pass


class RateLimitError(TwitterError):
    limit: int
    remaining: int
    reset: int

    def __init__(self, code: int, message: str, headers: CIMultiDictProxy[str]) -> None:
        self.code = code
        self.message = message
        self.limit = int(headers["x-rate-limit-limit"])
        # this is probably always zero, so no need to get it
        # self.remaining = int(headers["x-rate-limit-remaining"])
        self.reset = int(headers["x-rate-limit-reset"])
        time_till_reset = int(time.time() - self.reset)
        Exception.__init__(self, f"Rate limit exceeded. Will reset in {time_till_reset} seconds "
                                 f"(endpoint is limited to {self.limit} requests in 15 minutes)")


async def check_error(resp: ClientResponse) -> Any:
    try:
        resp_data = await resp.json()
    except ContentTypeError:
        resp.raise_for_status()
        return
    except json.JSONDecodeError:
        resp.raise_for_status()
        raise

    if not isinstance(resp_data, dict) or "errors" not in resp_data:
        resp.raise_for_status()
        return resp_data

    try:
        error = resp_data["errors"][0]
        code = error["code"]
        message = error["message"]
    except (KeyError, IndexError):
        resp.raise_for_status()
        raise
    if code == 88:
        raise RateLimitError(code, message, resp.headers)
    elif code == 32:
        raise TwitterAuthError(code, message)
    raise TwitterError(code, message)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.'
from typing import Dict, Type, TypeVar, List, Callable, Awaitable, Any

from mautrix.util.logging import TraceLogger

T = TypeVar('T')
Handler = Callable[[T], Awaitable[Any]]
HandlerMap = Dict[Type[T], List[Handler]]


class TwitterDispatcher:
    """
    This class is used to dispatch events that :class:`TwitterPoller` and :class:`TwitterStreamer`
    get from Twitter.
    """
    log: TraceLogger
    _handlers: HandlerMap

    async def dispatch(self, event: T) -> None:
        """
        Dispatch an event to handlers registered with :meth:`add_handler`.

        Args:
            event: The event to dispatch.
        """
        self.log.trace("Dispatching %s", event)
        for handler in self._handlers[type(event)]:
            try:
                await handler(event)
            except Exception:
                self.log.exception(f"Error while handling event of type {type(event)}")

    def add_handler(self, event_type: Type[T], handler: Handler) -> None:
        """
        Add an event handler.

        Args:
            event_type: The type of event to handle.
            handler: The handler function.
        """
        self._handlers[event_type].append(handler)

    def remove_handler(self, event_type: Type[T], handler: Handler) -> None:
        """
        Remove an event handler.

        Args:
            event_type: The type of event the handler was registered for.
            handler: The handler function to remove.
        """
        self._handlers[event_type].remove(handler)



# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Awaitable, Dict
import logging
import json

from aiohttp import web

from mautrix.types import UserID
from mautrix.util.logging import TraceLogger

from .. import user as u


class ProvisioningAPI:
    log: TraceLogger = logging.getLogger("mau.web.provisioning")
    app: web.Application

    def __init__(self, shared_secret: str) -> None:
        self.app = web.Application()
        self.shared_secret = shared_secret
        self.app.router.add_get("/api/whoami", self.status)
        self.app.router.add_options("/api/login", self.login_options)
        self.app.router.add_post("/api/login", self.login)
        self.app.router.add_post("/api/logout", self.logout)

    @property
    def _acao_headers(self) -> Dict[str, str]:
        return {
            "Access-Control-Allow-Origin": "*",
            "Access-Control-Allow-Headers": "Authorization, Content-Type",
            "Access-Control-Allow-Methods": "POST, OPTIONS",
        }

    @property
    def _headers(self) -> Dict[str, str]:
        return {
            **self._acao_headers,
            "Content-Type": "application/json",
        }

    async def login_options(self, _: web.Request) -> web.Response:
        return web.Response(status=200, headers=self._headers)

    def check_token(self, request: web.Request) -> Awaitable['u.User']:
        try:
            token = request.headers["Authorization"]
            token = token[len("Bearer "):]
        except KeyError:
            raise web.HTTPBadRequest(body='{"error": "Missing Authorization header"}',
                                     headers=self._headers)
        except IndexError:
            raise web.HTTPBadRequest(body='{"error": "Malformed Authorization header"}',
                                     headers=self._headers)
        if token != self.shared_secret:
            raise web.HTTPForbidden(body='{"error": "Invalid token"}', headers=self._headers)
        try:
            user_id = request.query["user_id"]
        except KeyError:
            raise web.HTTPBadRequest(body='{"error": "Missing user_id query param"}',
                                     headers=self._headers)

        return u.User.get_by_mxid(UserID(user_id))

    async def status(self, request: web.Request) -> web.Response:
        user = await self.check_token(request)
        data = {
            "permissions": user.permission_level,
            "mxid": user.mxid,
            "twitter": None,
        }
        if await user.is_logged_in():
            data["twitter"] = (await user.get_info()).serialize()
        return web.json_response(data, headers=self._acao_headers)

    async def login(self, request: web.Request) -> web.Response:
        user = await self.check_token(request)

        try:
            data = await request.json()
        except json.JSONDecodeError:
            raise web.HTTPBadRequest(body='{"error": "Malformed JSON"}', headers=self._headers)

        try:
            auth_token = data["auth_token"]
            csrf_token = data["csrf_token"]
        except KeyError:
            raise web.HTTPBadRequest(body='{"error": "Missing keys"}', headers=self._headers)

        try:
            await user.connect(auth_token=auth_token, csrf_token=csrf_token)
        except Exception:
            self.log.debug("Failed to log in", exc_info=True)
            raise web.HTTPUnauthorized(body='{"error": "Twitter authorization failed"}',
                                       headers=self._headers)
        return web.Response(body='{}', status=200, headers=self._headers)

    async def logout(self, request: web.Request) -> web.Response:
        user = await self.check_token(request)
        await user.logout()
        return web.json_response({}, headers=self._acao_headers)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import List, Optional
from datetime import datetime

from attr import dataclass

from mautrix.types import SerializableAttrs, SerializableEnum

from .util import StringTimestamp


class ConversationType(SerializableEnum):
    ONE_TO_ONE = "ONE_TO_ONE"
    GROUP_DM = "GROUP_DM"


class TimelineStatus(SerializableEnum):
    AT_END = "AT_END"
    HAS_MORE = "HAS_MORE"


@dataclass
class Participant(SerializableAttrs):
    user_id: str

    # This seems to be only for one-to-one chats
    last_read_event_id: Optional[str] = None

    # These seem to be only for group chats
    join_time: Optional[StringTimestamp] = None
    join_conversation_event_id: Optional[str] = None
    is_admin: Optional[bool] = None


@dataclass
class Conversation(SerializableAttrs):
    conversation_id: str
    type: ConversationType
    participants: List[Participant]
    notifications_disabled: bool
    mention_notifications_disabled: bool
    trusted: bool
    low_quality: bool

    sort_event_id: Optional[str] = None
    sort_timestamp: StringTimestamp = datetime.fromtimestamp(0)

    # These are present in some responses
    min_entry_id: Optional[str] = None
    max_entry_id: Optional[str] = None
    status: Optional[TimelineStatus] = None

    # This seems to be only for one-to-one chats
    read_only: Optional[bool] = None

    # These seem to be only for group chats
    create_time: Optional[StringTimestamp] = None
    created_by_user_id: Optional[str] = None
    name: Optional[str] = None
    last_read_event_id: Optional[str] = None



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional

from attr import dataclass

from mautrix.types import SerializableAttrs

from .message_entity import MessageEntities
from .message_attachment import MessageAttachment
from .conversation import Conversation
from .util import StringTimestamp


@dataclass
class MessageData(SerializableAttrs):
    id: str
    time: StringTimestamp
    sender_id: str
    text: str
    recipient_id: Optional[str] = None
    conversation_id: Optional[str] = None
    entities: Optional[MessageEntities] = None
    attachment: Optional[MessageAttachment] = None


@dataclass
class MessageEntry(SerializableAttrs):
    id: str
    time: StringTimestamp
    conversation_id: str
    message_data: MessageData
    affects_sort: Optional[bool] = None
    request_id: Optional[str] = None

    @property
    def sender_id(self) -> str:
        return self.message_data.sender_id

    conversation: Optional[Conversation] = None



# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, ClassVar, TYPE_CHECKING

from attr import dataclass

from mautrix.types import RoomID, EventID
from mautrix.util.async_db import Database

fake_db = Database("") if TYPE_CHECKING else None


@dataclass
class Message:
    db: ClassVar[Database] = fake_db

    mxid: EventID
    mx_room: RoomID
    twid: int
    receiver: int

    async def insert(self) -> None:
        q = "INSERT INTO message (mxid, mx_room, twid, receiver) VALUES ($1, $2, $3, $4)"
        await self.db.execute(q, self.mxid, self.mx_room, self.twid, self.receiver)

    async def delete(self) -> None:
        q = "DELETE FROM message WHERE twid=$1 AND receiver=$2"
        await self.db.execute(q, self.twid, self.receiver)

    @classmethod
    async def delete_all(cls, room_id: RoomID) -> None:
        await cls.db.execute("DELETE FROM message WHERE mx_room=$1", room_id)

    @classmethod
    async def get_by_mxid(cls, mxid: EventID, mx_room: RoomID) -> Optional['Message']:
        row = await cls.db.fetchrow("SELECT mxid, mx_room, twid, receiver "
                                    "FROM message WHERE mxid=$1 AND mx_room=$2", mxid, mx_room)
        if not row:
            return None
        return cls(**row)

    @classmethod
    async def get_by_twid(cls, twid: int, receiver: int = 0) -> Optional['Message']:
        row = await cls.db.fetchrow("SELECT mxid, mx_room, twid, receiver "
                                    "FROM message WHERE twid=$1 AND receiver=$2", twid, receiver)
        if not row:
            return None
        return cls(**row)



# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, ClassVar, List, TYPE_CHECKING

from attr import dataclass
import asyncpg

from mautrix.types import RoomID
from mautrix.util.async_db import Database

from mautwitdm.types import ConversationType

fake_db = Database("") if TYPE_CHECKING else None


@dataclass
class Portal:
    db: ClassVar[Database] = fake_db

    twid: str
    receiver: int
    conv_type: ConversationType
    other_user: Optional[int]
    mxid: Optional[RoomID]
    name: Optional[str]
    encrypted: bool

    async def insert(self) -> None:
        q = ("INSERT INTO portal (twid, receiver, conv_type, other_user, mxid, name, encrypted) "
             "VALUES ($1, $2, $3, $4, $5, $6, $7)")
        await self.db.execute(q, self.twid, self.receiver, self.conv_type.value, self.other_user,
                              self.mxid, self.name, self.encrypted)

    async def update(self) -> None:
        q = ("UPDATE portal SET conv_type=$3, other_user=$4, mxid=$5, name=$6, encrypted=$7 "
             "WHERE twid=$1 AND receiver=$2")
        await self.db.execute(q, self.twid, self.receiver, self.conv_type.value, self.other_user,
                              self.mxid, self.name, self.encrypted)

    @classmethod
    def _from_row(cls, row: asyncpg.Record) -> 'Portal':
        data = {**row}
        return cls(conv_type=ConversationType(data.pop("conv_type")), **data)

    @classmethod
    async def get_by_mxid(cls, mxid: RoomID) -> Optional['Portal']:
        q = ("SELECT twid, receiver, conv_type, other_user, mxid, name, encrypted "
             "FROM portal WHERE mxid=$1")
        row = await cls.db.fetchrow(q, mxid)
        if not row:
            return None
        return cls._from_row(row)

    @classmethod
    async def get_by_twid(cls, twid: str, receiver: int = 0) -> Optional['Portal']:
        q = ("SELECT twid, receiver, conv_type, other_user, mxid, name, encrypted "
             "FROM portal WHERE twid=$1 AND receiver=$2")
        row = await cls.db.fetchrow(q, twid, receiver)
        if not row:
            return None
        return cls._from_row(row)

    @classmethod
    async def find_private_chats_of(cls, receiver: int) -> List['Portal']:
        q = ("SELECT twid, receiver, conv_type, other_user, mxid, name, encrypted FROM portal "
             "WHERE receiver=$1 AND conv_type='ONE_TO_ONE'")
        rows = await cls.db.fetch(q, receiver)
        return [cls._from_row(row) for row in rows]

    @classmethod
    async def find_private_chats_with(cls, other_user: int) -> List['Portal']:
        q = ("SELECT twid, receiver, conv_type, other_user, mxid, name, encrypted FROM portal "
             "WHERE other_user=$1 AND conv_type='ONE_TO_ONE'")
        rows = await cls.db.fetch(q, other_user)
        return [cls._from_row(row) for row in rows]

    @classmethod
    async def all_with_room(cls) -> List['Portal']:
        q = ("SELECT twid, receiver, conv_type, other_user, mxid, name, encrypted FROM portal "
             'WHERE mxid IS NOT NULL')
        rows = await cls.db.fetch(q)
        return [cls._from_row(row) for row in rows]



# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import (Dict, Tuple, Optional, List, Deque, Set, Any, Union, AsyncGenerator,
                    Awaitable, NamedTuple, TYPE_CHECKING, cast)
from collections import deque
from datetime import datetime
import asyncio

from yarl import URL
import magic

from mautwitdm.types import (ConversationType, Conversation, MessageData, Participant, ReactionKey,
                             TimelineStatus, ReactionCreateEntry, ReactionDeleteEntry,
                             MessageEntry, VideoVariant)
from mautrix.appservice import AppService, IntentAPI
from mautrix.bridge import BasePortal, NotificationDisabler, async_getter_lock
from mautrix.types import (EventID, MessageEventContent, RoomID, EventType, MessageType,
                           MediaMessageEventContent, ImageInfo, VideoInfo, ThumbnailInfo,
                           ContentURI, EncryptedFile, TextMessageEventContent)
from mautrix.errors import MatrixError, MForbidden
from mautrix.util.message_send_checkpoint import MessageSendCheckpointStatus
from mautrix.util.simple_lock import SimpleLock

from .db import Portal as DBPortal, Message as DBMessage, Reaction as DBReaction
from .config import Config
from .formatter import twitter_to_matrix
from . import user as u, puppet as p, matrix as m

if TYPE_CHECKING:
    from .__main__ import TwitterBridge

try:
    from mautrix.crypto.attachments import encrypt_attachment, decrypt_attachment
except ImportError:
    encrypt_attachment = decrypt_attachment = None

StateBridge = EventType.find("m.bridge", EventType.Class.STATE)
StateHalfShotBridge = EventType.find("uk.half-shot.bridge", EventType.Class.STATE)
BackfillEntryTypes = Union[MessageEntry, ReactionCreateEntry, ReactionDeleteEntry]
ReuploadedMediaInfo = NamedTuple('ReuploadedMediaInfo', mxc=Optional[ContentURI],
                                 decryption_info=Optional[EncryptedFile],
                                 mime_type=str, file_name=str, size=int)


class Portal(DBPortal, BasePortal):
    by_mxid: Dict[RoomID, 'Portal'] = {}
    by_twid: Dict[Tuple[str, int], 'Portal'] = {}
    config: Config
    matrix: 'm.MatrixHandler'
    az: AppService
    private_chat_portal_meta: bool

    _main_intent: Optional[IntentAPI]
    _create_room_lock: asyncio.Lock
    backfill_lock: SimpleLock
    _msgid_dedup: Deque[int]
    _reqid_dedup: Set[str]
    _reaction_dedup: Deque[Tuple[int, int, ReactionKey]]

    _main_intent: IntentAPI
    _last_participant_update: Set[int]
    _reaction_lock: asyncio.Lock

    def __init__(self, twid: str, receiver: int, conv_type: ConversationType,
                 other_user: Optional[int] = None, mxid: Optional[RoomID] = None,
                 name: Optional[str] = None, encrypted: bool = False) -> None:
        super().__init__(twid, receiver, conv_type, other_user, mxid, name, encrypted)
        self._create_room_lock = asyncio.Lock()
        self.log = self.log.getChild(twid)
        self._msgid_dedup = deque(maxlen=100)
        self._reaction_dedup = deque(maxlen=100)
        self._reqid_dedup = set()
        self._last_participant_update = set()

        self.backfill_lock = SimpleLock("Waiting for backfilling to finish before handling %s",
                                        log=self.log)
        self._main_intent = None
        self._reaction_lock = asyncio.Lock()

    @property
    def is_direct(self) -> bool:
        return self.conv_type == ConversationType.ONE_TO_ONE

    @property
    def main_intent(self) -> IntentAPI:
        if not self._main_intent:
            raise ValueError("Portal must be postinit()ed before main_intent can be used")
        return self._main_intent

    @classmethod
    def init_cls(cls, bridge: 'TwitterBridge') -> None:
        cls.config = bridge.config
        cls.matrix = bridge.matrix
        cls.az = bridge.az
        cls.loop = bridge.loop
        BasePortal.bridge = bridge
        cls.private_chat_portal_meta = cls.config["bridge.private_chat_portal_meta"]
        NotificationDisabler.puppet_cls = p.Puppet
        NotificationDisabler.config_enabled = cls.config["bridge.backfill.disable_notifications"]

    # region Misc

    async def _send_delivery_receipt(self, event_id: EventID) -> None:
        if event_id and self.config["bridge.delivery_receipts"]:
            try:
                await self.az.intent.mark_read(self.mxid, event_id)
            except Exception:
                self.log.exception("Failed to send delivery receipt for %s", event_id)

    async def _upsert_reaction(self, existing: DBReaction, intent: IntentAPI, mxid: EventID,
                               message: DBMessage, sender: Union['u.User', 'p.Puppet'],
                               reaction: ReactionKey) -> None:
        if existing:
            self.log.debug(f"_upsert_reaction redacting {existing.mxid} and inserting {mxid}"
                           f" (message: {message.mxid})")
            await intent.redact(existing.mx_room, existing.mxid)
            await existing.edit(reaction=reaction, mxid=mxid, mx_room=message.mx_room)
        else:
            self.log.debug(f"_upsert_reaction inserting {mxid} (message: {message.mxid})")
            await DBReaction(mxid=mxid, mx_room=message.mx_room, tw_msgid=message.twid,
                             tw_receiver=self.receiver, tw_sender=sender.twid,
                             reaction=reaction).insert()

    # endregion
    # region Matrix event handling

    async def _send_error_notice(self, body: str, err: Exception):
        await self._send_message(
            self.main_intent,
            TextMessageEventContent(msgtype=MessageType.NOTICE, body=f"\u26a0 {body}: {err}"),
        )

    def _status_from_exception(self, e: Exception) -> MessageSendCheckpointStatus:
        if isinstance(e, NotImplementedError):
            return MessageSendCheckpointStatus.UNSUPPORTED
        return MessageSendCheckpointStatus.PERM_FAILURE

    async def handle_matrix_message(self, sender: 'u.User', message: MessageEventContent,
                                    event_id: EventID) -> None:
        if not sender.client:
            self.log.debug(f"Ignoring message {event_id} as user is not connected")
            return
        try:
            await self._handle_matrix_message(sender, message, event_id)
        except Exception as e:
            status = self._status_from_exception(e)
            sender.send_remote_checkpoint(
                status, event_id, self.mxid, EventType.ROOM_MESSAGE, message.msgtype, error=e,
            )
            await self._send_error_notice("Your message may not have been bridged", e)

    async def _handle_matrix_message(self, sender: 'u.User', message: MessageEventContent,
                                    event_id: EventID) -> None:
        request_id = str(sender.client.new_request_id())
        self._reqid_dedup.add(request_id)
        text = message.body
        media_id = None
        if message.msgtype == MessageType.EMOTE:
            text = f"/me {text}"
        elif message.msgtype.is_media:
            if message.file and decrypt_attachment:
                data = await self.main_intent.download_media(message.file.url)
                data = decrypt_attachment(data, message.file.key.key,
                                          message.file.hashes.get("sha256"), message.file.iv)
            else:
                data = await self.main_intent.download_media(message.url)
            mime_type = message.info.mimetype or magic.from_buffer(data, mime=True)
            upload_resp = await sender.client.upload(data, mime_type=mime_type)
            media_id = upload_resp.media_id
            text = ""
        resp = await sender.client.conversation(self.twid).send(text, media_id=media_id,
                                                                request_id=request_id)
        await self._send_delivery_receipt(event_id)
        sender.send_remote_checkpoint(
            status=MessageSendCheckpointStatus.SUCCESS,
            event_id=event_id,
            room_id=self.mxid,
            event_type=EventType.ROOM_MESSAGE,
            message_type=message.msgtype,
        )
        resp_msg_id = int(resp.entries[0].message.id)
        self._msgid_dedup.appendleft(resp_msg_id)
        msg = DBMessage(mxid=event_id, mx_room=self.mxid, twid=resp_msg_id, receiver=self.receiver)
        await msg.insert()
        self._reqid_dedup.remove(request_id)
        self.log.debug(f"Handled Matrix message {event_id} -> {resp_msg_id}")

    async def handle_matrix_reaction(self, sender: 'u.User', event_id: EventID,
                                     reacting_to: EventID, reaction_val: str) -> None:
        try:
            await self._handle_matrix_reaction(sender, event_id, reacting_to, reaction_val)
        except Exception as e:
            status = self._status_from_exception(e)
            self.log.exception(f"Failed to react to {event_id}")
            sender.send_remote_checkpoint(
                status, event_id, self.mxid, EventType.REACTION, error=e
            )
            await self._send_error_notice(f"Failed to react to {event_id}", e)

    async def _handle_matrix_reaction(self, sender: 'u.User', event_id: EventID,
                                     reacting_to: EventID, reaction_val: str) -> None:
        reaction = ReactionKey.from_emoji(reaction_val)
        message = await DBMessage.get_by_mxid(reacting_to, self.mxid)
        if not message:
            raise ValueError(f"Ignoring reaction to unknown event {reacting_to}")

        async with self._reaction_lock:
            dedup_id = (message.twid, sender.twid, reaction)
            self._reaction_dedup.appendleft(dedup_id)

            existing = await DBReaction.get_by_twid(message.twid, message.receiver, sender.twid)
            if existing and existing.reaction == reaction:
                return

            await sender.client.conversation(self.twid).react(message.twid, reaction)
            sender.send_remote_checkpoint(
                MessageSendCheckpointStatus.SUCCESS,
                event_id,
                self.mxid,
                EventType.REACTION,
            )
            await self._send_delivery_receipt(event_id)
            await self._upsert_reaction(existing, self.main_intent, event_id, message, sender,
                                        reaction)
            self.log.debug(f"{sender.mxid} reacted to {message.twid} with {reaction}")

    async def handle_matrix_redaction(self, sender: 'u.User', event_id: EventID,
                                      redaction_event_id: EventID) -> None:
        try:
            await self._handle_matrix_redaction(sender, event_id)
        except Exception as e:
            status = self._status_from_exception(e)
            sender.send_remote_checkpoint(
                status, event_id, self.mxid, EventType.ROOM_REDACTION, error=e
            )
            await self._send_error_notice(f"Failed to redact {event_id}", e)
        else:
            sender.send_remote_checkpoint(
                MessageSendCheckpointStatus.SUCCESS,
                redaction_event_id,
                self.mxid,
                EventType.ROOM_REDACTION,
            )
            await self._send_delivery_receipt(redaction_event_id)

    async def _handle_matrix_redaction(self, sender: 'u.User', event_id: EventID) -> None:
        assert self.mxid, "MXID is None"

        async with self._reaction_lock:
            reaction = await DBReaction.get_by_mxid(event_id, self.mxid)
            if reaction:
                try:
                    await reaction.delete()
                    await sender.client.conversation(self.twid).delete_reaction(reaction.tw_msgid,
                                                                                reaction.reaction)
                except Exception:
                    self.log.exception("Removing reaction failed")
                    raise
                else:
                    self.log.trace(f"Removed {reaction} after Matrix redaction")

                return

        raise NotImplementedError("Message redactions are not supported.")

    async def handle_matrix_leave(self, user: 'u.User') -> None:
        if self.is_direct:
            self.log.info(f"{user.mxid} left private chat portal with {self.twid}")
            if user.twid == self.receiver:
                self.log.info(f"{user.mxid} was the recipient of this portal. "
                              "Cleaning up and deleting...")
                await self.cleanup_and_delete()
        else:
            self.log.debug(f"{user.mxid} left portal to {self.twid}")
            # TODO cleanup if empty

    # endregion
    # region Twitter event handling

    async def handle_twitter_message(self, source: 'u.User', sender: 'p.Puppet',
                                     message: MessageData, request_id: str) -> None:
        msg_id = int(message.id)
        if request_id in self._reqid_dedup:
            self.log.debug(f"Ignoring message {message.id} by {message.sender_id}"
                           " as it was sent by us (request_id in dedup queue)")
        elif msg_id in self._msgid_dedup:
            self.log.debug(f"Ignoring message {message.id} by {message.sender_id}"
                           " as it was already handled (message.id in dedup queue)")
        elif await DBMessage.get_by_twid(msg_id, self.receiver) is not None:
            self.log.debug(f"Ignoring message {message.id} by {message.sender_id}"
                           " as it was already handled (message.id found in database)")
        else:
            self._msgid_dedup.appendleft(msg_id)
            intent = sender.intent_for(self)
            event_id = None
            if message.attachment:
                content = await self._handle_twitter_attachment(source, sender, message)
                if content:
                    event_id = await self._send_message(intent, content, timestamp=message.time)
            if message.text and not message.text.isspace():
                content = await twitter_to_matrix(message)
                event_id = await self._send_message(intent, content, timestamp=message.time)
            if event_id:
                msg = DBMessage(mxid=event_id, mx_room=self.mxid, twid=msg_id,
                                receiver=self.receiver)
                await msg.insert()
                await self._send_delivery_receipt(event_id)
                self.log.debug(f"Handled Twitter message {msg_id} -> {event_id}")

    @staticmethod
    def _is_better_mime(best: VideoVariant, current: VideoVariant) -> bool:
        order = ["video/mp4"]
        try:
            best_quality = order.index(best.content_type)
        except (IndexError, ValueError):
            best_quality = -1
        try:
            current_quality = order.index(current.content_type)
        except (IndexError, ValueError):
            current_quality = -1
        return current_quality > best_quality

    async def _handle_twitter_attachment(self, source: 'u.User', sender: 'p.Puppet',
                                         message: MessageData
                                         ) -> Optional[MediaMessageEventContent]:
        content = None
        intent = sender.intent_for(self)
        media = message.attachment.media
        if media:
            reuploaded_info = await self._reupload_twitter_media(source, media.media_url_https,
                                                                 intent)
            thumbnail_info = None
            if media.video_info:
                thumbnail_info = reuploaded_info
                best_variant = None
                for variant in media.video_info.variants:
                    if ((not best_variant or (variant.bitrate or 0) > (best_variant.bitrate or 0)
                         or self._is_better_mime(best_variant, variant))):
                        best_variant = variant
                reuploaded_info = await self._reupload_twitter_media(source, best_variant.url,
                                                                     intent)
            content = MediaMessageEventContent(body=reuploaded_info.file_name,
                                               url=reuploaded_info.mxc,
                                               file=reuploaded_info.decryption_info,
                                               external_url=media.media_url_https)
            if message.attachment.video:
                content.msgtype = MessageType.VIDEO
                content.info = VideoInfo(mimetype=reuploaded_info.mime_type,
                                         size=reuploaded_info.size,
                                         width=media.original_info.width,
                                         height=media.original_info.height,
                                         duration=media.video_info.duration_millis // 1000)
            elif message.attachment.photo or message.attachment.animated_gif:
                content.msgtype = MessageType.IMAGE
                content.info = ImageInfo(mimetype=reuploaded_info.mime_type,
                                         size=reuploaded_info.size,
                                         width=media.original_info.width,
                                         height=media.original_info.height)
            if thumbnail_info:
                content.info.thumbnail_url = thumbnail_info.mxc
                content.info.thumbnail_file = thumbnail_info.decryption_info
                content.info.thumbnail_info = ThumbnailInfo(mimetype=thumbnail_info.mime_type,
                                                            size=thumbnail_info.size,
                                                            width=media.original_info.width,
                                                            height=media.original_info.height)
            # Remove the attachment link from message.text
            start, end = media.indices
            message.text = message.text[:start] + message.text[end:]
        elif message.attachment.tweet:
            # TODO special handling for tweets?
            pass
        return content

    async def _reupload_twitter_media(self, source: 'u.User', url: str, intent: IntentAPI
                                      ) -> ReuploadedMediaInfo:
        file_name = URL(url).name
        data, mime_type = await source.client.download_media(url)

        upload_mime_type = mime_type
        upload_file_name = file_name
        decryption_info = None
        if self.encrypted and encrypt_attachment:
            data, decryption_info = encrypt_attachment(data)
            upload_mime_type = "application/octet-stream"
            upload_file_name = None

        mxc = await intent.upload_media(data, mime_type=upload_mime_type,
                                        filename=upload_file_name)

        if decryption_info:
            decryption_info.url = mxc
            mxc = None

        return ReuploadedMediaInfo(mxc, decryption_info, mime_type, file_name, len(data))

    async def handle_twitter_reaction_add(self, sender: 'p.Puppet', msg_id: int,
                                          reaction: ReactionKey, time: datetime) -> None:
        async with self._reaction_lock:
            dedup_id = (msg_id, sender.twid, reaction)
            if dedup_id in self._reaction_dedup:
                return
            self._reaction_dedup.appendleft(dedup_id)

        existing = await DBReaction.get_by_twid(msg_id, self.receiver, sender.twid)
        if existing and existing.reaction == reaction:
            return

        message = await DBMessage.get_by_twid(msg_id, self.receiver)
        if not message:
            self.log.debug(f"Ignoring reaction to unknown message {msg_id}")
            return

        intent = sender.intent_for(self)
        mxid = await intent.react(message.mx_room, message.mxid, reaction.emoji, timestamp=time)
        self.log.debug(f"{sender.twid} reacted to {message.mxid} -> {mxid}")
        await self._upsert_reaction(existing, intent, mxid, message, sender, reaction)

    async def handle_twitter_reaction_remove(self, sender: 'p.Puppet', msg_id: int,
                                             key: ReactionKey) -> None:
        reaction = await DBReaction.get_by_twid(msg_id, self.receiver, sender.twid)
        if reaction and reaction.reaction == key:
            try:
                await sender.intent_for(self).redact(reaction.mx_room, reaction.mxid)
            except MForbidden:
                await self.main_intent.redact(reaction.mx_room, reaction.mxid)
            await reaction.delete()
            self.log.debug(f"Removed {reaction} after Twitter removal")

    async def handle_twitter_receipt(self, sender: 'p.Puppet', read_up_to: int) -> None:
        message = await DBMessage.get_by_twid(read_up_to, self.receiver)
        if not message:
            self.log.debug(f"Ignoring read receipt from {sender.twid} "
                           f"up to unknown message {read_up_to}")
            return

        self.log.debug(f"{sender.twid} read messages up to {read_up_to} ({message.mxid})")
        await sender.intent_for(self).mark_read(message.mx_room, message.mxid)

    # endregion
    # region Updating portal info

    async def update_info(self, conv: Conversation) -> None:
        if self.conv_type == ConversationType.ONE_TO_ONE:
            if not self.other_user:
                participant = next(pcp for pcp in conv.participants
                                   if int(pcp.user_id) != self.receiver)
                self.other_user = int(participant.user_id)
                await self.update()
            puppet = await p.Puppet.get_by_twid(self.other_user)
            if not self._main_intent:
                self._main_intent = puppet.default_mxid_intent
            if self.encrypted or self.private_chat_portal_meta:
                changed = await self._update_name(puppet.name)
            else:
                changed = False
        else:
            changed = await self._update_name(conv.name)
        if changed:
            await self.update_bridge_info()
            await self.update()
        await self._update_participants(conv.participants)

    async def update_name(self, name: str) -> None:
        if not self.encrypted and not self.private_chat_portal_meta:
            return

        changed = await self._update_name(name)

        if changed:
            await self.update_bridge_info()
            await self.update()

    async def _update_name(self, name: str) -> bool:
        if self.name != name:
            self.name = name
            if self.mxid:
                await self.main_intent.set_room_name(self.mxid, name)
            return True
        return False

    async def _update_participants(self, participants: List[Participant]) -> None:
        if not self.mxid:
            return

        # Store the current member list to prevent unnecessary updates
        current_members = {int(participant.user_id) for participant in participants}
        if current_members == self._last_participant_update:
            self.log.trace("Not updating participants: list matches cached list")
            return
        self._last_participant_update = current_members

        # Make sure puppets who should be here are here
        for participant in participants:
            twid = int(participant.user_id)
            puppet = await p.Puppet.get_by_twid(twid)
            await puppet.intent_for(self).ensure_joined(self.mxid, bot=self.main_intent)
            if participant.last_read_event_id:
                await self.handle_twitter_receipt(puppet, int(participant.last_read_event_id))

        # Kick puppets who shouldn't be here
        for user_id in await self.main_intent.get_room_members(self.mxid):
            twid = p.Puppet.get_id_from_mxid(user_id)
            if twid and twid not in current_members:
                await self.main_intent.kick_user(self.mxid, p.Puppet.get_mxid_from_id(twid),
                                                 reason="User had left this Twitter chat")

    # endregion
    # region Backfilling

    async def backfill(self, source: 'u.User', is_initial: bool = False) -> None:
        if not is_initial:
            raise RuntimeError("Non-initial backfilling is not supported")
        limit = self.config["bridge.backfill.initial_limit"]
        if limit == 0:
            return
        elif limit < 0:
            limit = None
        with self.backfill_lock:
            await self._backfill(source, limit)

    async def _backfill(self, source: 'u.User', limit: int) -> None:
        self.log.debug("Backfilling history through %s", source.mxid)

        entries = await self._fetch_backfill_entries(source, limit)
        if not entries:
            self.log.debug("Didn't get any entries from server")
            return

        self.log.debug("Got %d entries from server", len(entries))

        backfill_leave = await self._invite_own_puppet_backfill(source)
        async with NotificationDisabler(self.mxid, source):
            for entry in reversed(entries):
                await self._handle_backfill_entry(source, entry)
        for intent in backfill_leave:
            self.log.trace("Leaving room with %s post-backfill", intent.mxid)
            await intent.leave_room(self.mxid)
        self.log.info("Backfilled %d messages through %s", len(entries), source.mxid)

    async def _fetch_backfill_entries(self, source: 'u.User', limit: int
                                      ) -> List[BackfillEntryTypes]:
        conv = source.client.conversation(self.twid)
        entries = []
        self.log.debug("Fetching up to %d messages through %s", limit, source.twid)
        try:
            max_id = None
            while True:
                resp = await conv.fetch(max_id=max_id, include_info=False)
                max_id = resp.min_entry_id
                for entry in resp.entries:
                    if entry:
                        entries += entry.all_types
                    if len(entries) >= limit:
                        break
                if len(entries) >= limit:
                    self.log.debug("Got more messages than limit")
                    break
                elif resp.status == TimelineStatus.AT_END:
                    self.log.debug("Got all messages in conversation")
                    break
        except Exception:
            self.log.warning("Exception while fetching messages", exc_info=True)

        return [entry for entry in entries
                if isinstance(entry, (MessageEntry, ReactionCreateEntry, ReactionDeleteEntry))]

    async def _invite_own_puppet_backfill(self, source: 'u.User') -> Set[IntentAPI]:
        backfill_leave = set()
        # TODO we should probably only invite the puppet when needed
        if self.config["bridge.backfill.invite_own_puppet"]:
            self.log.debug("Adding %s's default puppet to room for backfilling", source.mxid)
            sender = await p.Puppet.get_by_twid(source.twid)
            await self.main_intent.invite_user(self.mxid, sender.default_mxid)
            await sender.default_mxid_intent.join_room_by_id(self.mxid)
            backfill_leave.add(sender.default_mxid_intent)
        return backfill_leave

    async def _handle_backfill_entry(self, source: 'u.User', entry: BackfillEntryTypes) -> None:
        sender = await p.Puppet.get_by_twid(int(entry.sender_id))
        if isinstance(entry, MessageEntry):
            await self.handle_twitter_message(source, sender, entry.message_data, entry.request_id)
        if isinstance(entry, ReactionCreateEntry):
            await self.handle_twitter_reaction_add(sender, int(entry.message_id),
                                                   entry.reaction_key, entry.time)
        elif isinstance(entry, ReactionDeleteEntry):
            await self.handle_twitter_reaction_remove(sender, int(entry.message_id),
                                                      entry.reaction_key)

    # endregion
    # region Bridge info state event

    @property
    def bridge_info_state_key(self) -> str:
        return f"net.maunium.twitter://twitter/{self.twid}"

    @property
    def bridge_info(self) -> Dict[str, Any]:
        return {
            "bridgebot": self.az.bot_mxid,
            "creator": self.main_intent.mxid,
            "protocol": {
                "id": "twitter",
                "displayname": "Twitter DM",
                "avatar_url": self.config["appservice.bot_avatar"],
            },
            "channel": {
                "id": self.twid,
                "displayname": self.name,
            }
        }

    async def update_bridge_info(self) -> None:
        if not self.mxid:
            self.log.debug("Not updating bridge info: no Matrix room created")
            return
        try:
            self.log.debug("Updating bridge info...")
            await self.main_intent.send_state_event(self.mxid, StateBridge,
                                                    self.bridge_info, self.bridge_info_state_key)
            # TODO remove this once https://github.com/matrix-org/matrix-doc/pull/2346 is in spec
            await self.main_intent.send_state_event(self.mxid, StateHalfShotBridge,
                                                    self.bridge_info, self.bridge_info_state_key)
        except Exception:
            self.log.warning("Failed to update bridge info", exc_info=True)

    # endregion
    # region Creating Matrix rooms

    async def create_matrix_room(self, source: 'u.User', info: Conversation) -> Optional[RoomID]:
        if self.mxid:
            try:
                await self._update_matrix_room(source, info)
            except Exception:
                self.log.exception("Failed to update portal")
            return self.mxid
        async with self._create_room_lock:
            return await self._create_matrix_room(source, info)

    def _get_invite_content(self, double_puppet: Optional['p.Puppet']) -> Dict[str, Any]:
        invite_content = {}
        if double_puppet:
            invite_content["fi.mau.will_auto_accept"] = True
        if self.is_direct:
            invite_content["is_direct"] = True
        return invite_content

    async def _add_user(self, user: 'u.User') -> None:
        puppet = await p.Puppet.get_by_custom_mxid(user.mxid)
        await self.main_intent.invite_user(self.mxid, user.mxid, check_cache=True,
                                           extra_content=self._get_invite_content(puppet))
        if puppet:
            did_join = await puppet.intent.ensure_joined(self.mxid)
            if did_join and self.is_direct:
                await user.update_direct_chats({self.main_intent.mxid: [self.mxid]})

    async def _update_matrix_room(self, source: 'u.User', info: Conversation) -> None:
        await self._add_user(source)
        await self.update_info(info)

        # TODO
        # up = DBUserPortal.get(source.fbid, self.fbid, self.fb_receiver)
        # if not up:
        #     in_community = await source._community_helper.add_room(source._community_id, self.mxid)
        #     DBUserPortal(user=source.fbid, portal=self.fbid, portal_receiver=self.fb_receiver,
        #                  in_community=in_community).insert()
        # elif not up.in_community:
        #     in_community = await source._community_helper.add_room(source._community_id, self.mxid)
        #     up.edit(in_community=in_community)

    async def _create_matrix_room(self, source: 'u.User', info: Conversation) -> Optional[RoomID]:
        if self.mxid:
            await self._update_matrix_room(source, info)
            return self.mxid
        await self.update_info(info)
        self.log.debug("Creating Matrix room")
        name: Optional[str] = None
        initial_state = [{
            "type": str(StateBridge),
            "state_key": self.bridge_info_state_key,
            "content": self.bridge_info,
        }, {
            # TODO remove this once https://github.com/matrix-org/matrix-doc/pull/2346 is in spec
            "type": str(StateHalfShotBridge),
            "state_key": self.bridge_info_state_key,
            "content": self.bridge_info,
        }]
        invites = []
        if self.config["bridge.encryption.default"] and self.matrix.e2ee:
            self.encrypted = True
            initial_state.append({
                "type": "m.room.encryption",
                "content": {"algorithm": "m.megolm.v1.aes-sha2"},
            })
            if self.is_direct:
                invites.append(self.az.bot_mxid)
        if self.encrypted or self.private_chat_portal_meta or not self.is_direct:
            name = self.name
        if self.config["appservice.community_id"]:
            initial_state.append({
                "type": "m.room.related_groups",
                "content": {"groups": [self.config["appservice.community_id"]]},
            })

        # We lock backfill lock here so any messages that come between the room being created
        # and the initial backfill finishing wouldn't be bridged before the backfill messages.
        with self.backfill_lock:
            creation_content = {}
            if not self.config["bridge.federate_rooms"]:
                creation_content["m.federate"] = False
            self.mxid = await self.main_intent.create_room(
                name=name,
                is_direct=self.is_direct,
                initial_state=initial_state,
                invitees=invites,
                creation_content=creation_content,
            )
            if not self.mxid:
                raise Exception("Failed to create room: no mxid returned")

            if self.encrypted and self.matrix.e2ee and self.is_direct:
                try:
                    await self.az.intent.ensure_joined(self.mxid)
                except Exception:
                    self.log.warning("Failed to add bridge bot "
                                     f"to new private chat {self.mxid}")

            await self.update()
            self.log.debug(f"Matrix room created: {self.mxid}")
            self.by_mxid[self.mxid] = self
            await self._add_user(source)
            await self._update_participants(info.participants)

            puppet = await p.Puppet.get_by_custom_mxid(source.mxid)
            if puppet:
                try:
                    await puppet.intent.join_room_by_id(self.mxid)
                    if self.is_direct:
                        await source.update_direct_chats({self.main_intent.mxid: [self.mxid]})
                except MatrixError:
                    self.log.debug("Failed to join custom puppet into newly created portal",
                                   exc_info=True)

            if not info.trusted:
                msg = "This is a message request. Replying here will accept the request."
                if info.low_quality:
                    msg += " Note: Twitter has marked this as a \"low quality\" message."
                await self.main_intent.send_notice(self.mxid, msg)

            # TODO
            # in_community = await source._community_helper.add_room(source._community_id, self.mxid)
            # DBUserPortal(user=source.fbid, portal=self.fbid, portal_receiver=self.fb_receiver,
            #              in_community=in_community).upsert()

            try:
                await self.backfill(source, is_initial=True)
            except Exception:
                self.log.exception("Failed to backfill new portal")

            # Update participants again after backfill to sync read receipts
            self._last_participant_update = set()
            await self._update_participants(info.participants)

        return self.mxid

    # endregion
    # region Database getters

    async def postinit(self) -> None:
        self.by_twid[(self.twid, self.receiver)] = self
        if self.mxid:
            self.by_mxid[self.mxid] = self
        if self.other_user and self.is_direct:
            self._main_intent = (await p.Puppet.get_by_twid(self.other_user)).default_mxid_intent
        elif not self.is_direct:
            self._main_intent = self.az.intent

    async def delete(self) -> None:
        await DBMessage.delete_all(self.mxid)
        self.by_mxid.pop(self.mxid, None)
        self.mxid = None
        self.encrypted = False
        await self.update()

    async def save(self) -> None:
        await self.update()

    @classmethod
    def all_with_room(cls) -> AsyncGenerator['Portal', None]:
        return cls._db_to_portals(super().all_with_room())

    @classmethod
    def find_private_chats_with(cls, other_user: int) -> AsyncGenerator['Portal', None]:
        return cls._db_to_portals(super().find_private_chats_with(other_user))

    @classmethod
    async def _db_to_portals(cls, query: Awaitable[List['Portal']]) -> AsyncGenerator['Portal', None]:
        portals = await query
        for index, portal in enumerate(portals):
            try:
                yield cls.by_twid[(portal.twid, portal.receiver)]
            except KeyError:
                await portal.postinit()
                yield portal

    @classmethod
    @async_getter_lock
    async def get_by_mxid(cls, mxid: RoomID) -> Optional['Portal']:
        try:
            return cls.by_mxid[mxid]
        except KeyError:
            pass

        portal = cast(cls, await super().get_by_mxid(mxid))
        if portal is not None:
            await portal.postinit()
            return portal

        return None

    @classmethod
    @async_getter_lock
    async def get_by_twid(cls, twid: str, *, receiver: int = 0,
                          conv_type: Optional[ConversationType] = None) -> Optional['Portal']:
        if conv_type == ConversationType.GROUP_DM and receiver != 0:
            receiver = 0
        try:
            return cls.by_twid[(twid, receiver)]
        except KeyError:
            pass

        portal = cast(cls, await super().get_by_twid(twid, receiver))
        if portal is not None:
            await portal.postinit()
            return portal

        if conv_type is not None:
            portal = cls(twid, receiver, conv_type)
            await portal.insert()
            await portal.postinit()
            return portal

        return None

    # endregion



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, Optional

from attr import dataclass

from mautrix.types import SerializableAttrs, SerializableEnum

from .util import StringTimestamp
from .conversation import Conversation


class ReactionKey(SerializableEnum):
    FUNNY = "funny"  # 😂
    SURPRISED = "surprised"  # 😲
    SAD = "sad"  # 😢
    LIKE = "like"  # ❤
    EXCITED = "excited"  # 🔥
    AGREE = "agree"  # 👍
    DISAGREE = "disagree"  # 👎

    @property
    def emoji(self) -> str:
        return _key_to_emoji[self]

    @classmethod
    def from_emoji(cls, emoji: str) -> 'ReactionKey':
        try:
            return _emoji_to_key[emoji.rstrip("\uFE0F")]
        except KeyError:
            raise NotImplementedError(f"Unsupported reaction emoji {emoji}")


_key_to_emoji: Dict[ReactionKey, str] = {
    ReactionKey.FUNNY: "\U0001F602",
    ReactionKey.SURPRISED: "\U0001F632",
    ReactionKey.SAD: "\U0001F622",
    ReactionKey.LIKE: "\u2764\uFE0F",
    ReactionKey.EXCITED: "\U0001F525",
    ReactionKey.AGREE: "\U0001F44D\uFE0F",
    ReactionKey.DISAGREE: "\U0001F44E\uFE0F",
}

_emoji_to_key: Dict[str, ReactionKey] = {
    "\U0001F44D": ReactionKey.AGREE,
    "\U0001F44E": ReactionKey.DISAGREE,
    "\u2764": ReactionKey.LIKE,
    "\U0001F525": ReactionKey.EXCITED,
    "\U0001F622": ReactionKey.SAD,
    "\U0001F632": ReactionKey.SURPRISED,
    "\U0001F62E": ReactionKey.SURPRISED,  # This is 😮, which is commonly used for surprised
    "\U0001F602": ReactionKey.FUNNY,
}


@dataclass
class ReactionCreateEntry(SerializableAttrs):
    id: str
    time: StringTimestamp
    conversation_id: str
    message_id: str
    reaction_key: ReactionKey
    sender_id: str
    affects_sort: Optional[bool] = None

    conversation: Optional[Conversation] = None


@dataclass
class ReactionDeleteEntry(SerializableAttrs):
    id: str
    time: StringTimestamp
    conversation_id: str
    message_id: str
    reaction_key: ReactionKey
    sender_id: str
    affects_sort: Optional[bool] = None

    conversation: Optional[Conversation] = None



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, Optional, Type, TypeVar, List, Callable, Awaitable, Any, Union
import logging
import asyncio
import time

from attr import dataclass
from aiohttp import ClientSession
from yarl import URL

from .types import PollResponse, InitialStateResponse
from .errors import check_error, RateLimitError
from .conversation import Conversation
from .dispatcher import TwitterDispatcher

T = TypeVar('T')
Handler = Callable[[T], Awaitable[Any]]
HandlerMap = Dict[Type[T], List[Handler]]


class PollingStarted:
    pass


class PollingStopped:
    pass


@dataclass
class PollingErrored:
    error: Exception
    fatal: bool
    count: int


class PollingErrorResolved:
    pass


class TwitterPoller(TwitterDispatcher):
    """This class handles polling for new messages using ``/dm/user_updates.json``."""
    dm_url: URL

    log: logging.Logger
    loop: asyncio.AbstractEventLoop
    http: ClientSession
    headers: Dict[str, str]
    skip_poll_wait: asyncio.Event

    poll_sleep: int = 3
    error_sleep: int = 5
    max_poll_errors: int = 12
    poll_cursor: Optional[str]
    dispatch_initial_resp: bool
    _poll_task: Optional[asyncio.Task]
    _typing_in: Optional[Conversation]

    @property
    def poll_params(self) -> Dict[str, str]:
        return {
            "cards_platform": "Web-12",
            "include_cards": "1",
            "include_ext_alt_text": "true",
            "include_quote_count": "true",
            "include_reply_count": "1",
            "tweet_mode": "extended",
            "dm_users": "false",
            "include_groups": "true",
            "include_inbox_timelines": "true",
            "include_ext_media_color": "true",
            "supports_reactions": "true",
            "ext": "mediaColor,altText,mediaStats,highlightedLabel",
        }

    @property
    def full_state_params(self) -> Dict[str, str]:
        return {
            "include_profile_interstitial_type": "1",
            "include_blocking": "1",
            "include_blocked_by": "1",
            "include_followed_by": "1",
            "include_want_retweets": "1",
            "include_mute_edge": "1",
            "include_can_dm": "1",
            "include_can_media_tag": "1",
            "skip_status": "1",
            **self.poll_params,
        }

    async def inbox_initial_state(self, set_poll_cursor: bool = True) -> InitialStateResponse:
        """
        Get the initial DM inbox state, including conversations, user profiles and some messages.

        This also gets the initial :attr:`poll_cursor` value.

        Returns:
            The response data from the server.
        """
        url = (self.dm_url / "inbox_initial_state.json").with_query({
            **self.full_state_params,
            "filter_low_quality": "true",
            "include_quality": "all",
        })
        async with self.http.get(url, headers=self.headers) as resp:
            data = await check_error(resp)
            response = InitialStateResponse.deserialize(data["inbox_initial_state"])
            if set_poll_cursor:
                self.poll_cursor = response.cursor
            return response

    async def _poll_once(self) -> PollResponse:
        if not self.poll_cursor:
            raise RuntimeError("Cursor must be set to poll")
        url = (self.dm_url / "user_updates.json").with_query({
            **self.poll_params,
            "cursor": self.poll_cursor,
            "filter_low_quality": "true",
            "include_quality": "all",
        })
        async with self.http.get(url, headers=self.headers) as resp:
            data = await check_error(resp)
            try:
                user_events = data["user_events"]
            except KeyError:
                try:
                    inbox_initial_state = data["inbox_initial_state"]
                except KeyError:
                    self.log.warning("Got unknown poll response: %s", data)
                    raise
                else:
                    response = InitialStateResponse.deserialize(inbox_initial_state)
            else:
                response = PollResponse.deserialize(user_events)
            self.poll_cursor = response.cursor
            return response

    async def poll_forever(self, raise_exceptions: bool = True) -> None:
        """
        Poll for updates forever.

        Args:
            raise_exceptions: Whether or not fatal errors should be raised after logging.
                :class:`asyncio.CancelledError` will not be raised in any case.
        """
        try:
            await self._poll_forever()
        except asyncio.CancelledError:
            self.log.debug("Polling stopped")
            await self.dispatch(PollingStopped())
        except Exception as e:
            await self.dispatch(PollingErrored(e, fatal=True, count=0))
            self.log.exception("Fatal error while polling")
            if raise_exceptions:
                raise

    async def dispatch_all(self, resp: Union[PollResponse, InitialStateResponse]) -> None:
        for user in (resp.users or {}).values():
            await self.dispatch(user)
        for conversation in (resp.conversations or {}).values():
            await self.dispatch(conversation)
        for entry in resp.entries or []:
            if not entry:
                continue
            for entry_type in entry.all_types:
                try:
                    entry_type.conversation = resp.conversations[entry_type.conversation_id]
                except KeyError:
                    msg = ("Poll response didn't contain conversation info "
                           f"for {entry_type.conversation_id} "
                           f"(entry type {type(entry_type)}, ID {entry_type.id}")
                    if entry_type.conversation:
                        self.log.debug(f"{msg}, but the entry had its own conversation info")
                        await self.dispatch(entry_type)
                    else:
                        # TODO we could still dispatch the event and guess the conversation type
                        #      in event handlers if necessary.
                        self.log.warning(msg)
                else:
                    await self.dispatch(entry_type)

    async def _poll_forever(self) -> None:
        if not self.poll_cursor:
            self.log.debug("Poll cursor not set, calling initial state to get cursor")
            resp = await self.inbox_initial_state()
            if self.dispatch_initial_resp:
                await self.dispatch_all(resp)
        await self.dispatch(PollingStarted())
        errors = 0
        while True:
            try:
                resp = await self._poll_once()
            except RateLimitError as e:
                sleep = e.reset - int(time.time())
                self.log.warning(f"Got rate limit until {e.reset} while polling, "
                                 f"waiting {sleep} seconds")
                if self.poll_sleep < 8:
                    self.poll_sleep += 1
                    self.log.debug(f"Increased poll sleep to {self.poll_sleep}")
                await asyncio.sleep(sleep)
                continue
            except Exception as e:
                if errors > self.max_poll_errors > 0:
                    self.log.debug(f"Error count ({errors}) exceeded maximum, "
                                   f"raising error as fatal")
                    raise
                errors += 1
                sleep = min(15 * 60, self.error_sleep * errors)
                self.log.warning(f"Error while polling, retrying in {sleep}s", exc_info=True)
                await self.dispatch(PollingErrored(e, fatal=False, count=errors))
                await asyncio.sleep(sleep)
                continue
            if errors > 0:
                errors = 0
                await self.dispatch(PollingErrorResolved())
            await self.dispatch_all(resp)
            try:
                await asyncio.wait_for(self.skip_poll_wait.wait(), self.poll_sleep)
            except asyncio.TimeoutError:
                pass
            if self._typing_in:
                await self._typing_in.mark_typing()

    def is_polling(self) -> bool:
        return self._poll_task and not self._poll_task.done()

    def start_polling(self) -> asyncio.Task:
        """
        Start polling forever in the background. This calls :meth:`poll_forever` and puts it in an
        asyncio Task. The task is stored so it can be cancelled with :meth:`stop_polling`.

        Returns:
            The created asyncio task.
        """
        self.stop_polling()
        self.log.debug("Starting poll task")
        self._poll_task = asyncio.create_task(self.poll_forever())
        return self._poll_task

    def stop_polling(self) -> None:
        """Stop the ongoing poll task. Any ongoing handlers will also be cancelled."""
        if self._poll_task:
            self.log.debug("Cancelling ongoing poll task")
            self._poll_task.cancel()
            self._poll_task = None



# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, ClassVar, List, TYPE_CHECKING

from attr import dataclass
from yarl import URL
import asyncpg

from mautrix.types import UserID, SyncToken, ContentURI
from mautrix.util.async_db import Database

fake_db = Database("") if TYPE_CHECKING else None


@dataclass
class Puppet:
    db: ClassVar[Database] = fake_db

    twid: int
    name: Optional[str]
    photo_url: Optional[str]
    photo_mxc: Optional[ContentURI]

    is_registered: bool

    custom_mxid: Optional[UserID]
    access_token: Optional[str]
    next_batch: Optional[SyncToken]
    base_url: Optional[URL]

    async def insert(self) -> None:
        q = ("INSERT INTO puppet (twid, name, photo_url, photo_mxc, is_registered, custom_mxid,"
             "                    access_token, next_batch, base_url) "
             "VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9)")
        await self.db.execute(q, self.twid, self.name, self.photo_url, self.photo_mxc,
                              self.is_registered, self.custom_mxid, self.access_token,
                              self.next_batch, str(self.base_url) if self.base_url else None)

    async def update(self) -> None:
        q = ("UPDATE puppet SET name=$2, photo_url=$3, photo_mxc=$4, is_registered=$5,"
             "                  custom_mxid=$6, access_token=$7, next_batch=$8, base_url=$9 "
             "WHERE twid=$1")
        await self.db.execute(q, self.twid, self.name, self.photo_url, self.photo_mxc,
                              self.is_registered, self.custom_mxid, self.access_token,
                              self.next_batch, str(self.base_url) if self.base_url else None)

    @classmethod
    def _from_row(cls, row: asyncpg.Record) -> 'Puppet':
        data = {**row}
        base_url_str = data.pop("base_url")
        base_url = URL(base_url_str) if base_url_str is not None else None
        return cls(base_url=base_url, **data)

    @classmethod
    async def get_by_twid(cls, twid: int) -> Optional['Puppet']:
        row = await cls.db.fetchrow("SELECT twid, name, photo_url, photo_mxc, is_registered,"
                                    "       custom_mxid, access_token, next_batch, base_url "
                                    "FROM puppet WHERE twid=$1", twid)
        if not row:
            return None
        return cls._from_row(row)

    @classmethod
    async def get_by_custom_mxid(cls, mxid: UserID) -> Optional['Puppet']:
        row = await cls.db.fetchrow("SELECT twid, name, photo_url, photo_mxc, is_registered,"
                                    "       custom_mxid, access_token, next_batch, base_url "
                                    "FROM puppet WHERE custom_mxid=$1", mxid)
        if not row:
            return None
        return cls._from_row(row)

    @classmethod
    async def all_with_custom_mxid(cls) -> List['Puppet']:
        rows = await cls.db.fetch("SELECT twid, name, photo_url, photo_mxc, is_registered,"
                                  "       custom_mxid, access_token, next_batch, base_url "
                                  "FROM puppet WHERE custom_mxid IS NOT NULL")
        return [cls._from_row(row) for row in rows]



from mautrix.util.async_db import Database

from .upgrade import upgrade_table
from .user import User
from .puppet import Puppet
from .portal import Portal
from .message import Message
from .reaction import Reaction


def init(db: Database) -> None:
    for table in (User, Puppet, Portal, Message, Reaction):
        table.db = db


__all__ = ["upgrade_table", "User", "Puppet", "Portal", "Message", "Reaction"]



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional

from attr import dataclass

from mautrix.types import SerializableAttrs

from .util import StringDateTime


@dataclass
class User(SerializableAttrs):
    id: int
    id_str: str
    name: str
    screen_name: str
    profile_image_url: Optional[str]
    profile_image_url_https: Optional[str]
    description: str
    created_at: StringDateTime
    verified: bool
    protected: bool

    can_media_tag: bool
    following: bool
    follow_request_sent: bool

    friends_count: int
    followers_count: int

    statuses_count: Optional[int] = None

    followed_by: Optional[bool] = None
    blocking: Optional[bool] = None
    suspended: Optional[bool] = None
    bocked_by: Optional[bool] = None

    url: Optional[str] = None
    location: Optional[str] = None



# mautrix-twitter - A Matrix-Twitter DM puppeting bridge
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, Dict, AsyncIterable, Awaitable, AsyncGenerator, TYPE_CHECKING, cast
from os import path

from aiohttp import ClientSession
from yarl import URL

from mautwitdm.types import User
from mautrix.bridge import BasePuppet, async_getter_lock
from mautrix.appservice import IntentAPI
from mautrix.types import ContentURI, UserID, SyncToken, RoomID
from mautrix.util.simple_template import SimpleTemplate

from .db import Puppet as DBPuppet
from .config import Config
from . import portal as p

if TYPE_CHECKING:
    from .__main__ import TwitterBridge


class Puppet(DBPuppet, BasePuppet):
    by_twid: Dict[int, 'Puppet'] = {}
    by_custom_mxid: Dict[UserID, 'Puppet'] = {}
    hs_domain: str
    mxid_template: SimpleTemplate[int]

    config: Config

    default_mxid_intent: IntentAPI
    default_mxid: UserID

    def __init__(self, twid: int, name: Optional[str] = None, photo_url: Optional[str] = None,
                 photo_mxc: Optional[ContentURI] = None, is_registered: bool = False,
                 custom_mxid: Optional[UserID] = None, access_token: Optional[str] = None,
                 next_batch: Optional[SyncToken] = None, base_url: Optional[URL] = None) -> None:
        super().__init__(twid=twid, name=name, photo_url=photo_url, photo_mxc=photo_mxc,
                         is_registered=is_registered, custom_mxid=custom_mxid,
                         access_token=access_token, next_batch=next_batch, base_url=base_url)
        self.log = self.log.getChild(str(twid))

        self.default_mxid = self.get_mxid_from_id(twid)
        self.default_mxid_intent = self.az.intent.user(self.default_mxid)
        self.intent = self._fresh_intent()

    @classmethod
    def init_cls(cls, bridge: 'TwitterBridge') -> AsyncIterable[Awaitable[None]]:
        cls.config = bridge.config
        cls.loop = bridge.loop
        cls.mx = bridge.matrix
        cls.az = bridge.az
        cls.hs_domain = cls.config["homeserver.domain"]
        cls.mxid_template = SimpleTemplate(cls.config["bridge.username_template"], "userid",
                                           prefix="@", suffix=f":{cls.hs_domain}", type=int)
        cls.sync_with_custom_puppets = cls.config["bridge.sync_with_custom_puppets"]
        cls.homeserver_url_map = {server: URL(url) for server, url
                                  in cls.config["bridge.double_puppet_server_map"].items()}
        cls.allow_discover_url = cls.config["bridge.double_puppet_allow_discovery"]
        cls.login_shared_secret_map = {server: secret.encode("utf-8") for server, secret
                                       in cls.config["bridge.login_shared_secret_map"].items()}
        cls.login_device_name = "Twitter DM Bridge"
        return (puppet.try_start() async for puppet in cls.all_with_custom_mxid())

    def intent_for(self, portal: 'p.Portal') -> IntentAPI:
        if portal.other_user == self.twid or (self.config["bridge.backfill.invite_own_puppet"]
                                              and portal.backfill_lock.locked):
            return self.default_mxid_intent
        return self.intent

    async def update_info(self, info: User) -> None:
        update = False
        update = await self._update_name(info) or update
        update = await self._update_avatar(info.profile_image_url_https) or update
        if update:
            await self.update()

    @classmethod
    def _get_displayname(cls, info: User) -> str:
        return cls.config["bridge.displayname_template"].format(displayname=info.name, id=info.id,
                                                                username=info.screen_name)

    async def _update_name(self, info: User) -> bool:
        name = self._get_displayname(info)
        if name != self.name:
            self.name = name
            await self.default_mxid_intent.set_displayname(self.name)
            await self._update_portal_names()
            return True
        return False

    async def _update_portal_names(self) -> None:
        async for portal in p.Portal.find_private_chats_with(self.twid):
            await portal.update_name(self.name)

    async def _update_avatar(self, image_url: str) -> bool:
        if image_url != self.photo_url:
            self.photo_url = image_url
            url = URL(self.photo_url.replace("_normal.", "_400x400."))
            file_name = path.basename(url.path)
            async with ClientSession() as sess, sess.get(url) as resp:
                content_type = resp.headers["Content-Type"]
                resp_data = await resp.read()
            mxc = await self.default_mxid_intent.upload_media(data=resp_data, filename=file_name,
                                                              mime_type=content_type)
            self.photo_mxc = mxc
            await self.default_mxid_intent.set_avatar_url(mxc)
            return True
        return False

    async def default_puppet_should_leave_room(self, room_id: RoomID) -> bool:
        portal = await p.Portal.get_by_mxid(room_id)
        return portal and portal.other_user != self.twid

    # region Database getters

    def _add_to_cache(self) -> None:
        self.by_twid[self.twid] = self
        if self.custom_mxid:
            self.by_custom_mxid[self.custom_mxid] = self

    async def save(self) -> None:
        await self.update()

    @classmethod
    async def get_by_mxid(cls, mxid: UserID, create: bool = True) -> Optional['Puppet']:
        twid = cls.get_id_from_mxid(mxid)
        if twid:
            return await cls.get_by_twid(twid, create=create)
        return None

    @classmethod
    @async_getter_lock
    async def get_by_custom_mxid(cls, mxid: UserID) -> Optional['Puppet']:
        try:
            return cls.by_custom_mxid[mxid]
        except KeyError:
            pass

        puppet = cast(cls, await super().get_by_custom_mxid(mxid))
        if puppet:
            puppet._add_to_cache()
            return puppet

        return None

    @classmethod
    def get_id_from_mxid(cls, mxid: UserID) -> Optional[int]:
        return cls.mxid_template.parse(mxid)

    @classmethod
    def get_mxid_from_id(cls, twid: int) -> UserID:
        return UserID(cls.mxid_template.format_full(twid))

    @classmethod
    @async_getter_lock
    async def get_by_twid(cls, twid: int, *, create: bool = True) -> Optional['Puppet']:
        try:
            return cls.by_twid[twid]
        except KeyError:
            pass

        puppet = cast(cls, await super().get_by_twid(twid))
        if puppet is not None:
            puppet._add_to_cache()
            return puppet

        if create:
            puppet = cls(twid)
            await puppet.insert()
            puppet._add_to_cache()
            return puppet

        return None

    @classmethod
    async def all_with_custom_mxid(cls) -> AsyncGenerator['Puppet', None]:
        puppets = await super().all_with_custom_mxid()
        puppet: cls
        for index, puppet in enumerate(puppets):
            try:
                yield cls.by_twid[puppet.twid]
            except KeyError:
                puppet._add_to_cache()
                yield puppet

    # endregion



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, Union, List

from attr import dataclass

from mautrix.types import SerializableAttrs

from .message import MessageEntry
from .conversation import Conversation
from .reaction import ReactionDeleteEntry, ReactionCreateEntry
from .util import StringTimestamp


@dataclass
class TrustConversationEntry(SerializableAttrs):
    id: str
    time: StringTimestamp
    conversation_id: str
    reason: str
    affects_sort: Optional[bool] = None

    conversation: Optional[Conversation] = None


@dataclass
class ConversationCreateEntry(SerializableAttrs):
    id: str
    time: StringTimestamp
    conversation_id: str
    affects_sort: Optional[bool] = None
    request_id: Optional[str] = None

    conversation: Optional[Conversation] = None


@dataclass
class ConversationNameUpdateEntry(SerializableAttrs):
    id: str
    time: StringTimestamp
    conversation_id: str
    conversation_name: str
    by_user_id: str
    affects_sort: Optional[bool] = None

    conversation: Optional[Conversation] = None


@dataclass
class ConversationReadEntry(SerializableAttrs):
    id: str
    last_read_event_id: str
    time: StringTimestamp
    conversation_id: str
    affects_sort: Optional[bool] = None

    conversation: Optional[Conversation] = None


EntryType = Union[MessageEntry, TrustConversationEntry, ConversationCreateEntry,
                  ConversationNameUpdateEntry, ReactionCreateEntry, ReactionCreateEntry,
                  ConversationReadEntry]


@dataclass
class Entry(SerializableAttrs):
    message: Optional[MessageEntry] = None
    trust_conversation: Optional[TrustConversationEntry] = None
    conversation_create: Optional[ConversationCreateEntry] = None
    conversation_name_update: Optional[ConversationNameUpdateEntry] = None
    reaction_create: Optional[ReactionCreateEntry] = None
    reaction_delete: Optional[ReactionDeleteEntry] = None
    conversation_read: Optional[ConversationReadEntry] = None

    @property
    def all_types(self) -> List[EntryType]:
        items = (self.conversation_create, self.conversation_name_update, self.trust_conversation,
                 self.message, self.reaction_create, self.reaction_delete, self.conversation_read)
        return [item for item in items
                if item is not None]
